/* vi: set sw=4 ts=4: */
/*
 * Mini ps implementation(s) for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include "busybox.h"
#ifdef CONFIG_SELINUX
#include <fs_secure.h>
#include <ss.h>
#include <flask_util.h>          /* for is_flask_enabled() */
#endif

static const int TERMINAL_WIDTH = 79;      /* not 80 in case terminal has linefold bug */

enum PSFormatIndex {
    PS_FMT_PID,
    PS_FMT_PPID,
    PS_FMT_SZ,
    PS_FMT_RSS,
    PS_FMT_DATA,
    PS_FMT_SHR,
    PS_FMT_MAJFL,
    PS_FMT_MINFL,
    PS_FMT_STAT,
    PS_FMT_TIME,
    PS_FMT_COMM,
    PS_FMT_CONTEXT,
    PS_FMT_UID,
    PS_FMT_NICE,
    PS_FMT_USER,
    PS_FMT_PRIORITY,
    PS_FMT_PGRP,
    PS_FMT_SESSION,
};

typedef struct PSFormat {
    const char *name;
    const char *disp;
} PSFormat;

static const PSFormat ps_formats[] = {
    { "pid",  "  PID", },
    { "ppid", " PPID", },
    { "size", "   SZ", }, /* XXX: check man ps */
    { "rss",  "  RSS", },
    { "data", " DATA", }, /* XXX: check man ps */
    { "shr",    "  SHR", }, /* XXX: check man ps */
    { "majfl",  "MAJFL", }, /* XXX: check man ps */
    { "minfl",  "MINFL", }, /* XXX: check man ps */
    { "stat",   "STAT", },
    { "time",   "  TIME", },
    { "comm",   "COMMAND", },
    { "context", "Context                         " },
    { "uid", "  UID", },
    { "nice", " NI", },
    { "user", "USER    ", },
    { "priority", " PRI", },
    { "pgrp", " PGRP", },
    { "session", " SESS", },
    { NULL, NULL, },
};

static int find_format(const char **pp)
{
    const PSFormat *psf;
    const char *p;
    char buf[128], *q;

    p = *pp;
    q = buf;
    while (*p != '\0' && *p != ',') {
        if ((q - buf) < (int)sizeof(buf) - 1)
            *q++ = *p;
        p++;
    }
    *q = '\0';
    if (*p == ',')
        p++;
    *pp = p;

    for(psf = ps_formats; psf->name != NULL; psf++) {
        if (!strcmp(psf->name, buf)) {
            return psf - ps_formats;
        }
    }
    return -1;
}

static void show_ps_formats(void)
{
    const PSFormat *psf;

    printf("\nAvailable ps formats:");
    for(psf = ps_formats; psf->name != NULL; psf++) {
        printf(" %s", psf->name);
    }
    printf("\n");
}

extern int ps_main(int argc, char **argv)
{
	procps_status_t * p;
	int i, len;
	int terminal_width = TERMINAL_WIDTH;
        int wide = 0;
        int opt;
        const char *ps_fmt_str, *fmt_ptr;
        enum PSFormatIndex fmt_index;
        
#ifdef CONFIG_SELINUX
	int use_selinux = 0;
	security_id_t sid;
	if(is_flask_enabled() && argv[1] && !strcmp(argv[1], "-c") )
		use_selinux = 1;
#endif
        ps_fmt_str = "pid,ppid,stat,time,comm";
#ifdef CONFIG_SELINUX
        if (use_selinux) 
            ps_fmt_str = "pid,uid,size,stat,comm";
#endif
        while ((opt = getopt (argc, argv, "wmo:")) != -1) {
            switch (opt) {
            case 'w':
                wide = 1;
                break;
                
            case 'm':
                ps_fmt_str = "pid,ppid,size,rss,data,shr,majfl,minfl,stat,time,comm";
                break;

            case 'o':
                ps_fmt_str = optarg;
                break;
            default:
                bb_show_usage();
                break;
            }
        }
	get_terminal_width_height(0, &terminal_width, NULL);
	/* Go one less... */
	terminal_width--;

        fmt_ptr = ps_fmt_str;
        while (*fmt_ptr != '\0') {
            fmt_index = find_format(&fmt_ptr);
            if (fmt_index == (enum PSFormatIndex)-1) {
            invalid_format:
                show_ps_formats();
                bb_show_usage();
                return 1;
            }
            if (fmt_ptr != ps_fmt_str)
                printf(" ");
            printf("%s", ps_formats[fmt_index].disp);
        }
        printf("\n");

#ifdef CONFIG_SELINUX
	while ((p = procps_scan(1, use_selinux, &sid)) != 0) {
#else
	while ((p = procps_scan(1)) != 0) {
#endif
                fmt_ptr = ps_fmt_str;
                len = 0;
                while (*fmt_ptr != '\0') {
                    fmt_index = find_format(&fmt_ptr);
                    if (fmt_index == (enum PSFormatIndex)-1)
                        goto invalid_format;
                    if (fmt_ptr != ps_fmt_str)
                        len += printf(" ");
                    switch(fmt_index) {
                    case PS_FMT_PID:
                        len += printf("%5d", p->pid);
                        break;
                    case PS_FMT_PPID:
                        len += printf("%5d", p->ppid);
                        break;
                    case PS_FMT_SZ:
                        len += printf("%5ld", p->vms);
                        break;
                    case PS_FMT_RSS:
                        len += printf("%5ld", p->rss);
                        break;
                    case PS_FMT_DATA:
                        len += printf("%5ld", p->vds);
                        break;
                    case PS_FMT_SHR:
                        len += printf("%5ld", p->vss);
                        break;
                    case PS_FMT_MAJFL:
                        len += printf("%5u", p->majflt);
                        break;
                    case PS_FMT_MINFL:
                        len += printf("%5u", p->minflt);
                        break;
                    case PS_FMT_STAT:
                        len += printf("%c%c%c ", 
                                      p->state[0], p->state[1], p->state[2]);
                        break;
                    case PS_FMT_TIME:
                        {
                            unsigned long t;
                            t = (p->utime + p->stime) / HZ;
                            len += printf("%3lu:%02lu", t / 60, t % 60);
                        }
                        break;
                    case PS_FMT_COMM:
                        {
                            char *namecmd = p->cmd;
                            i = terminal_width-len;
                            if(namecmd != 0 && namecmd[0] != 0) {
                                if(i < 0)
                                    i = 0;
                                if((int)strlen(namecmd) > i && !wide)
                                    namecmd[i] = 0;
                                len += printf("%s", namecmd);
                            } else {
                                namecmd = p->short_cmd;
                                if(i < 2)
                                    i = 2;
                                if((int)strlen(namecmd) > (i-2))
                                    namecmd[i-2] = 0;
                                len += printf("[%s]", namecmd);
                            }
                        }
                        break;
                    case PS_FMT_CONTEXT:
#ifdef CONFIG_SELINUX
                        {
                            char sbuf[128];
                            len = sizeof(sbuf);
                            if(security_sid_to_context(sid, (security_context_t)&sbuf, &len))
				strcpy(sbuf, "unknown");
                            
                            len += printf("%-32s", sbuf);
                        }
#else
                        len += printf("%-32s", "unknown");
#endif
                        break;
                    case PS_FMT_UID:
                        len += printf("%5d", p->uid);
                        break;
                    case PS_FMT_NICE:
                        len += printf("%3d", p->nice);
                        break;
                    case PS_FMT_USER:
                        len += printf("%-8s", p->user);
                        break;
                    case PS_FMT_PRIORITY:
                        len += printf("%4d", p->priority);
                        break;
                    case PS_FMT_PGRP:
                        len += printf("%5d", p->pgrp);
                        break;
                    case PS_FMT_SESSION:
                        len += printf("%5d", p->session);
                        break;
                    }
                }
                printf("\n");
		free(p->cmd);
	}
	return EXIT_SUCCESS;
}

