#define _XOPEN_SOURCE /* glibc2 needs this for strptime */
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <limits.h>
#include <time.h>
#include <sys/stat.h>

#include <tar.h>

//#define DEBUG_TAR 1
//#define DEBUG_DIR_STRUCT 1
//#define DEBUG_PARSE 1

#ifdef DEBUG_TAR
#define DTPRINTF(fmt, args...) do { printf(fmt , ##args); } while (0)
#else
#define DTPRINTF(fmt, args...) do { } while (0)
#endif

#ifdef DEBUG_DIR_STRUCT
#define DSPRINTF(fmt, args...) do { printf(fmt , ##args); } while (0)
#else
#define DSPRINTF(fmt, args...) do { } while (0)
#endif

#ifdef DEBUG_PARSE
#define DPPRINTF(fmt, args...) do { printf(fmt , ##args); } while (0)
#else
#define DPPRINTF(fmt, args...) do { } while (0)
#endif

typedef struct file_t file_t;
struct file_t {
    const unsigned char *name; /* Name in archive */
    int mode;                  /* Permissions     */
    int uid;
    int gid;
    int ctime;
    const unsigned char *uname;
    const unsigned char *gname;
    /* The following information are not mandatory for all file types */
    const unsigned char *local_name;
    const unsigned char *link_name;
    int major;
    int minor;
};

/* Tar file creation */
#define USTAR_BLOC_SIZE 512

static uint32_t tar_chk (uint32_t crc, const uint8_t *buf, int len)
{
    const uint8_t *p;
    int i;

    p = buf;
    for (i = 0; i < len; i++)
        crc += *p++;

    return crc;
}

static int emit_tar_data (FILE *fout, FILE *fin, int len)
{
    uint8_t bloc_buf[USTAR_BLOC_SIZE];
    int rem, l;

    rem = len;
    fseek(fin, 0, SEEK_SET);
    for (; rem != 0; rem -= l) {
        l = rem;
        if (l > USTAR_BLOC_SIZE)
            l = USTAR_BLOC_SIZE;
        if (fread(bloc_buf, l, 1, fin) != 1)
            break;
        if (l < USTAR_BLOC_SIZE)
            memset(bloc_buf + l, 0, USTAR_BLOC_SIZE - l);
        if (fwrite(bloc_buf, USTAR_BLOC_SIZE, 1, fout) != 1)
            break;
    }
    fclose(fin);

    return rem == 0 ? 0 : -1;
}

static int emit_tar_entry (FILE *f, file_t *file, int type)
{
    uint8_t bloc_buf[USTAR_BLOC_SIZE], *p;
    FILE *fin;
    uint32_t crc;
    int flen;

    fin = NULL;
    memset(bloc_buf, 0, USTAR_BLOC_SIZE);
    p = (unsigned char *)bloc_buf;
    /* Filename */
    strncpy(p, file->name, 100);
    p += 100;
    /* mode, UID, GID, size, time and type */
    switch (type) {
    case REGTYPE:
    case AREGTYPE:
    case CONTTYPE:
        /* Regular file */
        fin = fopen(file->local_name, "r");
        if (fin == NULL)
            return -1;
        fseek(fin, 0, SEEK_END);
        flen = ftell(fin);
        break;
    default:
        flen = 0;
        break;
    }
    sprintf(p, "%07o", file->mode);
    p += 8;
    sprintf(p, "%07o", file->uid);
    p += 8;
    sprintf(p, "%07o", file->gid);
    p += 8;
    sprintf(p, "%011o", flen);
    p += 12;
    sprintf(p, "%011o", file->ctime);
    p += 12;
    sprintf(p, "        %c", type);
    p += 9;
    /* linkname when relevant */
    switch (type) {
    case LNKTYPE:
    case SYMTYPE:
        /* link */
        strncpy(p, file->link_name, 100);
        break;
    default:
        break;
    }
    p += 100;
    /* TAR magic */
    strcpy(p, TMAGIC);
    p += TMAGLEN;
    /* TAR version */
    memcpy(p, TVERSION, TVERSLEN);
    p += TVERSLEN;
    /* uname */
    strncpy(p, file->uname, 32);
    p += 32;
    /* gname */
    strncpy(p, file->gname, 32);
    p += 32;
    switch (type) {
    case CHRTYPE:
    case BLKTYPE:
        /* Device */
        sprintf(p, "%07o\0%07o\0", file->major, file->minor);
        break;
    default:
        break;
    }
    /* Compute the header's checksum */
    crc = 0x00000000;
    crc = tar_chk(crc, bloc_buf, USTAR_BLOC_SIZE);
    /* Patch the checksum */
    p = bloc_buf + 148;
    sprintf(p, "%06o\0", crc & 0x1FFFF);
    /* Write the header */
    if (fwrite(bloc_buf, USTAR_BLOC_SIZE, 1, f) != 1)
        return -1;
    switch (type) {
    case REGTYPE:
    case AREGTYPE:
    case CONTTYPE:
        /* Copy content of the regular file */
        if (emit_tar_data(f, fin, flen) < 0)
            return -1;
#if 0
        /* Add an empty trailer bloc */
        memset(bloc_buf, 0, USTAR_BLOC_SIZE);
        if (fwrite(bloc_buf, USTAR_BLOC_SIZE, 1, f) != 1)
            return -1;
#endif
        break;
    default:
        break;
    }

    return 0;
}

static int emit_tar_trailer (FILE *f)
{
#if 0
    uint8_t bloc_buf[USTAR_BLOC_SIZE];

    memset(bloc_buf, 0, USTAR_BLOC_SIZE);
    if (fwrite(bloc_buf, USTAR_BLOC_SIZE, 2, f) != 2)
        return -1;
#endif

    return 0;
}

/* We keep a list of existing directories */
typedef struct dir_t dir_t;
struct dir_t {
    file_t file;
    int namelen;
    dir_t *children;
    dir_t *next;
};

static dir_t fake_dir = {
    .file      = {
        .name  = "",
        .mode  = 0755,
        .uid   = 0,
        .gid   = 0,
        .ctime = 0,
        .uname = "root",
        .gname = "root",
    },
    .namelen   = 0,
    .children  = NULL,
    .next      = NULL,
};

static dir_t *add_direntry (FILE *f, dir_t *parent, file_t *file, int dlen)
{
    unsigned char tmpname[PATH_MAX];
    dir_t **dirp, *dir;
    const unsigned char *p, *q;
    int len, sl, cl;

    DSPRINTF("%s: parent '%s' %d dir '%s' %d\n", __func__,
             parent->file.name, parent->namelen, file->name, dlen);
    len = parent->namelen;
    for (p = file->name + len; *p == '/'; p++)
        continue;
    q = strchr(p, '/');
    if (q == NULL) {
        cl = strlen(file->name);
        sl = -1;
    } else {
        cl = q - file->name;
        sl = cl;
    }
    DSPRINTF("%s: cl %d sl %d\n", __func__, cl, sl);
    for (dirp = &parent->children; *dirp != NULL; dirp = &(*dirp)->next) {
        dir = *dirp;
        len = dir->namelen;
        DSPRINTF("Compare '%s' %d <=> '%s' '%s' %d\n",
               dir->file.name, len, file->name, p, cl);
        if (len == cl && memcmp(dir->file.name, file->name, len) == 0) {
            if (sl != -1) {
                /* Recurse into this child */
                DSPRINTF("%s: recurse into '%s' %d\n", __func__,
                         dir->file.name, dir->namelen);
                dir = add_direntry(f, dir, file, dlen);
                return dir;
            }
            /* This directory is already in the tree */
            DSPRINTF("%s: found '%s' %d\n", __func__,
                     dir->file.name, dir->namelen);
            return dir;
        }
    }
    /* We did not found the directory */
    if (sl == -1) {
        /* Create the leaf */
        len = dlen;
        strcpy(tmpname, file->name);
    } else {
        /* Create the missing intermediary directory */
        len = cl;
        memcpy(tmpname, file->name, len);
        tmpname[len] = '\0';
    }
    DSPRINTF("%s: create '%s' %d\n", __func__, tmpname, len);
    dir = malloc(sizeof(dir_t));
    if (dir == NULL)
        return NULL;
    memcpy(&dir->file, file, sizeof(file_t));
    dir->file.name = strdup(tmpname);
    if (dir->file.name == NULL) {
        free(dir);
        return NULL;
    }
    dir->namelen = len;
    dir->children = NULL;
    dir->next = NULL;
    *dirp = dir;
    /* Create the directory in the tar archive */
    if (emit_tar_entry(f, &dir->file, DIRTYPE) < 0)
        return NULL;
    if (sl != -1) {
        /* We must recurse to the last leaf */
        return add_direntry(f, parent, file, dlen);
    }

    return dir;
}

#ifdef DEBUG_DIR_STRUCT
static void dump_direntries (dir_t *parent)
{
    dir_t *dir;

    for (dir = parent->children; dir != NULL; dir = dir->next) {
        printf("entry '%s' %d\n", dir->file.name, dir->namelen);
        if (dir->children != NULL)
            dump_direntries(dir);
    }
}
#endif

static int add_dir (FILE *f, const unsigned char *dirname,
                    int mode, int uid, int gid, int ctime,
                    const unsigned char *uname, const unsigned char *gname)
{
    file_t tmpfile;

    tmpfile.name  = dirname;
    tmpfile.mode  = mode;
    tmpfile.uid   = uid;
    tmpfile.gid   = gid;
    tmpfile.ctime = ctime;
    tmpfile.uname = uname;
    tmpfile.gname = gname;
    if (add_direntry(f, &fake_dir, &tmpfile, strlen(dirname)) == NULL)
        return -1;

    return 0;
}

static int add_file (FILE *f, const unsigned char *filename, int dir_mode,
                     int mode, int uid, int gid, int ctime,
                     const unsigned char *uname, const unsigned char *gname,
                     const unsigned char *local_name)
{
    unsigned char tmpname[PATH_MAX];
    file_t tmpfile;
    const unsigned char *p;
    int len;

    tmpfile.uid        = uid;
    tmpfile.gid        = gid;
    tmpfile.ctime      = ctime;
    tmpfile.uname      = uname;
    tmpfile.gname      = gname;
    p = strrchr(filename, '/');
    if (p != NULL) {
        /* Create the parent directory */
        len = p - filename;
        memcpy(tmpname, filename, len);
        tmpname[len] = '\0';
        tmpfile.mode       = dir_mode;
        tmpfile.name       = tmpname;
        if (add_direntry(f, &fake_dir, &tmpfile, len) == NULL)
            return -1;
    }
    tmpfile.mode       = mode;
    tmpfile.name       = filename;
    tmpfile.local_name = local_name;
    /* Create the file in the tar archive */
    if (emit_tar_entry(f, &tmpfile, REGTYPE) < 0)
        return -1;

    return 0;
}

static int add_link (FILE *f, const unsigned char *name, int dir_mode,
                     int mode, int uid, int gid, int ctime,
                     const unsigned char *uname, const unsigned char *gname,
                     const unsigned char *link_name)
{
    unsigned char tmpname[PATH_MAX];
    file_t tmpfile;
    const unsigned char *p;
    int len;

    tmpfile.uid       = uid;
    tmpfile.gid       = gid;
    tmpfile.ctime     = ctime;
    tmpfile.uname     = uname;
    tmpfile.gname     = gname;
    p = strrchr(name, '/');
    if (p != NULL) {
        /* Create the parent directory */
        len = p - name;
        memcpy(tmpname, name, len);
        tmpname[len] = '\0';
        tmpfile.mode       = dir_mode;
        tmpfile.name       = tmpname;
        if (add_direntry(f, &fake_dir, &tmpfile, len) == NULL)
            return -1;
    }
    tmpfile.mode       = mode;
    tmpfile.name       = name;
    tmpfile.link_name = link_name;
    /* Create the link in the tar archive */
    if (emit_tar_entry(f, &tmpfile, LNKTYPE) < 0)
        return -1;

    return 0;
}

static int add_symlink (FILE *f, const unsigned char *name, int dir_mode,
                        int mode, int uid, int gid, int ctime,
                        const unsigned char *uname, const unsigned char *gname,
                        const unsigned char *link_name)
{
    unsigned char tmpname[PATH_MAX];
    file_t tmpfile;
    const unsigned char *p;
    int len;

    tmpfile.uid       = uid;
    tmpfile.gid       = gid;
    tmpfile.ctime     = ctime;
    tmpfile.uname     = uname;
    tmpfile.gname     = gname;
    p = strrchr(name, '/');
    if (p != NULL) {
        /* Create the parent directory */
        len = p - name;
        memcpy(tmpname, name, len);
        tmpname[len] = '\0';
        tmpfile.mode       = dir_mode;
        tmpfile.name       = tmpname;
        if (add_direntry(f, &fake_dir, &tmpfile, len) == NULL)
            return -1;
    }
    tmpfile.mode       = mode;
    tmpfile.name       = name;
    tmpfile.link_name = link_name;
    /* Create the symbolic link in the tar archive */
    if (emit_tar_entry(f, &tmpfile, SYMTYPE) < 0)
        return -1;

    return 0;
}

static int add_dev (FILE *f, const unsigned char *name, int dir_mode,
                    int mode, int uid, int gid, int ctime,
                    const unsigned char *uname, const unsigned char *gname,
                    int type, int major, int minor)
{
    unsigned char tmpname[PATH_MAX];
    file_t tmpfile;
    const unsigned char *p;
    int len;

    tmpfile.uid       = uid;
    tmpfile.gid       = gid;
    tmpfile.ctime     = ctime;
    tmpfile.uname     = uname;
    tmpfile.gname     = gname;
    p = strrchr(name, '/');
    if (p != NULL) {
        /* Create the parent directory */
        len = p - name;
        memcpy(tmpname, name, len);
        tmpname[len] = '\0';
        tmpfile.mode       = dir_mode;
        tmpfile.name       = tmpname;
        if (add_direntry(f, &fake_dir, &tmpfile, len) == NULL)
            return -1;
    }
    tmpfile.mode       = mode;
    tmpfile.name       = name;
    tmpfile.major     = major;
    tmpfile.minor     = minor;
    /* Create the device entry in the tar archive */
    if (emit_tar_entry(f, &tmpfile, type) < 0)
        return -1;

    return 0;
}

static inline add_blkdev (FILE *f, const unsigned char *name, int dir_mode,
                          int mode, int uid, int gid, int ctime,
                          const unsigned char *uname,
                          const unsigned char *gname,
                          int major, int minor)
{
    return add_dev(f, name, dir_mode, mode, uid, gid, ctime, uname, gname,
                   BLKTYPE, major, minor);
}

static inline add_chrdev (FILE *f, const unsigned char *name, int dir_mode,
                          int mode, int uid, int gid, int ctime,
                          const unsigned char *uname,
                          const unsigned char *gname,
                          int major, int minor)
{
    return add_dev(f, name, dir_mode, mode, uid, gid, ctime, uname, gname,
                   CHRTYPE, major, minor);
}

#if 0
static int add_fifo (FILE *f, const unsigned char *name, int dir_mode,
                     int mode, int uid, int gid, int ctime,
                     const unsigned char *uname, const unsigned char *gname)
{
    unsigned char tmpname[PATH_MAX];
    file_t tmpfile;
    const unsigned char *p;
    int len;

    tmpfile.uid       = uid;
    tmpfile.gid       = gid;
    tmpfile.ctime     = ctime;
    tmpfile.uname     = uname;
    tmpfile.gname     = gname;
    p = strrchr(name, '/');
    if (p != NULL) {
        /* Create the parent directory */
        len = p - name;
        memcpy(tmpname, name, len);
        tmpname[len] = '\0';
        tmpfile.mode       = dir_mode;
        tmpfile.name       = tmpname;
        if (add_direntry(f, &fake_dir, &tmpfile, len) == NULL)
            return -1;
    }
    tmpfile.mode       = mode;
    tmpfile.name       = name;
    /* Create the device entry in the tar archive */
    if (emit_tar_entry(f, &tmpfile, FIFOTYPE) < 0)
        return -1;

    return 0;
}
#endif

/* Read a line and break it up in tokens */
#define MAX_TOKENS          10
#define LINE_LEN           200

static const unsigned char *mount_point = "";
static int lct;

static int tokenize_line (FILE *f, unsigned char line[LINE_LEN],
                          unsigned char *parse[MAX_TOKENS])
{
    unsigned char *p, *q;
    int k, l, d, toks;

    p = line;
    q = p;
    l = 0;
    parse[0] = p;
    toks = 1;
    for (k = 0; k < LINE_LEN; k++) {
        for (;;) {
            d = fgetc(f);
            if (toks == 1 && isspace(d) && d != '\n')
                continue;
            toks = 0;
            if (d != '#')
                break;
            /* we handle only line comments */
            for (; d != EOF && d != '\n';)
                d = fgetc(f);
            if (d == '\n')
                lct++;
        }
        if (d == EOF) {
            fprintf(stderr, "Unexpected end-of-file\n");
            return -1;
        }
        if (d == '\n' || d == ' ' || d == '\t') {
            if (p != q) {
                *p++ = '\0';
                parse[l++] = q;
                DPPRINTF("New token '%s'\n", parse[l - 1]);
            }
            if (d == '\n') {
                lct++;
                break;
            }
            if (l == MAX_TOKENS) {
                fprintf(stderr, "Too many tokens\n");
                return -1;
            }
            q = p;
            toks = 1;
        } else {
            *p = d;
            p++;
        }
    }
    if (k == LINE_LEN) {
        fprintf(stderr, "Line too long\n");
        return -1;
    }
    for (k = l; k < MAX_TOKENS; k++)
        parse[l] = NULL;

    return l;
}

static int parse_proto (const unsigned char *proto_name,
                        const unsigned char *out_name,
                        int default_dir_mode, int ctime)
{
    unsigned char line[LINE_LEN], *token[MAX_TOKENS], *p, *q;
    FILE *f, *fout;
    int mount_point_len, nt, c, mode, uid, gid, major, minor;

    f = fopen(proto_name, "r");
    if (f == NULL) {
        fprintf(stderr, "Unable to open proto file '%s'\n", proto_name);
        return -1;
    }
    fout = fopen(out_name, "w");
    if (fout == NULL) {
        fprintf(stderr, "Unable to create outfile '%s'\n", out_name);
        return -1;
    }
    lct = 0;
    mount_point_len = strlen(mount_point);
    /* Skip first line (header) */
    nt = tokenize_line(f, line, token);
    if (nt < 0) {
        fprintf(stderr, "Unable to read header\n");
        return -1;
    }
    /* Skip bloc and inodes count */
    nt = tokenize_line(f, line, token);
    if (nt < 0) {
        fprintf(stderr, "Unable to read counts\n");
        return -1;
    }
    for (;;) {
        nt = tokenize_line(f, line, token);
        if (nt < 0) {
            fprintf(stderr, "Unable to parse line\n");
            return -1;
        }
#if 0 // XXX: allow empty lines
        if (nt == 0)
            break;
#else
        if (nt == 0)
            continue;
#endif
        p = token[0];
        if (*p == '\0')
            continue;
        if (*p == '$')
            break;
        if (nt == 3 && p[0] == 'd')
            continue;
        /* Check for initial directory as part of file system */
        if (memcmp(mount_point, p, mount_point_len) != 0 ||
            p[mount_point_len] != '/') {
            fprintf(stderr,
                    "warning: file '%s' not part of file-system '%s' %d %d\n",
                    p, mount_point, lct, nt);
            return -1;
        } else {
            p += mount_point_len + 1;
        }
        if (nt < 4) {
            /* we're in trouble */
            fprintf(stderr, "Missing tokens for '%s'\n", p);
            return -1;
        }
        q = token[1];
        /* Get mode and flags */
        mode = 0;
        switch (q[1]) {
        case 'u':
            mode |= TSUID;
            break;
        case '-':
            break;
        default:
            fprintf(stderr, "Invalid SUID flag for '%s'\n", p);
            return -1;
        }
        switch (q[2]) {
        case 'g':
            mode |= TSGID;
            break;
        case '-':
            break;
        default:
            fprintf(stderr, "Invalid GUID flag for '%s'\n", p);
            return -1;
        }
        switch (q[3]) {
        case 't':
            mode |= TSVTX;
            break;
        case '-':
            break;
        default:
            fprintf(stderr, "Invalid sticky flag for '%s'\n", p);
            return -1;
        }
        c = q[4] - '0';
        if (c < 0 || c > 7) {
            fprintf(stderr, "Invalid user perms %c for '%s'\n", q[4], p);
            return -1;
        }
        if (c & 4)
            mode |= TUREAD;
        if (c & 2)
            mode |= TUWRITE;
        if (c & 1)
            mode |= TUEXEC;
        c = q[5] - '0';
        if (c < 0 || c > 7) {
            fprintf(stderr, "Invalid group perms %c for '%s'\n", q[4], p);
            return -1;
        }
        if (c & 4)
            mode |= TGREAD;
        if (c & 2)
            mode |= TGWRITE;
        if (c & 1)
            mode |= TGEXEC;
        c = q[6] - '0';
        if (c < 0 || c > 7) {
            fprintf(stderr, "Invalid other perms %c for '%s'\n", q[4], p);
            return -1;
        }
        if (c & 4)
            mode |= TOREAD;
        if (c & 2)
            mode |= TOWRITE;
        if (c & 1)
            mode |= TOEXEC;
        uid = atoi(token[2]);
        gid = atoi(token[3]);
        switch (*q) {
        case '-':
        case 'C':
            /* Regular file */
            if (nt != 5) {
                fprintf(stderr, "Bad number of tokens for file '%s' %d\n", p, nt);
                return -1;
            }
            if (add_file(fout, p, default_dir_mode, mode, uid, gid, ctime,
                         "", "", token[4]) < 0) {
                fprintf(stderr, "Unable to add file '%s' '%s'\n", p, token[4]);
                return -1;
            }
            break;
        case 'd':
            /* Directory */
            if (nt != 4) {
                fprintf(stderr, "Bad number of tokens for dir '%s'\n", p);
                return -1;
            }
            c = strlen(p);
            if (p[c - 1] == '/')
                p[c - 1] = '\0';
            if (add_dir(fout, p, mode, uid, gid, ctime, "", "") < 0) {
                fprintf(stderr, "Unable to add dir '%s'\n", p);
                return -1;
            }
            break;
        case 'L':
            /* Hard link */
            if (nt != 5) {
                fprintf(stderr, "Bad number of tokens for link '%s'\n", p);
                return -1;
            }
            if (add_link(fout, p, default_dir_mode, mode, uid, gid, ctime,
                         "", "", token[4]) < 0) {
                fprintf(stderr, "Unable to add hard link '%s'\n", p);
                return -1;
            }
            break;
        case 'l':
            /* Symbolic link */
            if (nt != 5) {
                fprintf(stderr, "Bad number of tokens for symlink '%s'\n", p);
                return -1;
            }
            if (add_symlink(fout, p, default_dir_mode, mode, uid, gid, ctime,
                            "", "", token[4]) < 0) {
                fprintf(stderr, "Unable to add symlink '%s'\n", p);
                return -1;
            }
            break;
        case 'c':
            /* Char device */
            /* no break here */
        case 'b':
            /* Bloc device */
            if (nt < 5) {
                fprintf(stderr, "Missing token for device '%s'\n", p);
                return -1;
            }
            major = atoi(token[4]);
            minor = atoi(token[5]);
            if (*q == 'c') {
                if (add_chrdev(fout, p, default_dir_mode, mode, uid, gid,
                               ctime, "", "", major, minor) < 0) {
                    fprintf(stderr, "Unable to add char device '%s'\n", p);
                    return -1;
                }
            } else {
                if (add_blkdev(fout, p, default_dir_mode, mode, uid, gid,
                               ctime, "", "", major, minor) < 0) {
                    fprintf(stderr, "Unable to add bloc device '%s'\n", p);
                    return -1;
                }
            }
            break;
        default:
            fprintf(stderr, "Unknown file type '%c' for '%s'\n", *q, p);
            return -1;
        }
            
    }
    emit_tar_trailer(fout);
    fclose(fout);
    fclose(f);

    return 0;
}

static void usage (void)
{
    fprintf(stderr, "Usage: mktar [-h] [-v] [-d] [-r <time_str>] "
            "[-m <mount_point>] <proto_file> <tar_file>\n");
    fprintf(stderr, "\t-h: print this help\n");
    fprintf(stderr, "\t-v: increase verbosity\n");
    fprintf(stderr, "\t-d: change files creation time to program date\n");
    fprintf(stderr, "\t-r: use <time_str> as files creation time\n");
    fprintf(stderr, "\t-m: use <mount_point> as reference for all files\n");
}

int main (int argc, char **argv)
{
    struct tm tm;
    struct stat statbuf;
    const unsigned char *proto_name, *outname;
    int ctime, bin_time, default_dir_perms;
    int c, verbose;

    verbose = 0;
    default_dir_perms = 0755;
    ctime = time(NULL);
    stat(argv[0], &statbuf);
    bin_time = statbuf.st_mtime; /* time when binary was last modified */
    for (;;) {
        c = getopt(argc, argv, "dr:m:vh");
        if (c < 0)
            break;
        switch (c) {
        case 'd':
            ctime = bin_time;
            break;
        case 'r':
            /* Use reference time given as a string */
            if (strptime(optarg, "%a %b %d %H:%M:%S UTC %Y", &tm) != NULL)
                ctime = mktime(&tm);
            break;
        case 'm':
            mount_point = optarg;
            break;
        case 'v':
            verbose++;
            break;
        case 'h':
            usage();
            return 0;
        default:
            usage();
            fprintf(stderr, "Unknown argument\n");
            return 1;
        }
    }
    if (argc - optind != 2) {
        usage();
        fprintf(stderr, "Bad number of arguments\n");
        return 1;
    }
    proto_name = argv[optind++];
    outname = argv[optind];
    if (parse_proto(proto_name, outname, default_dir_perms, ctime) < 0)
        return 1;
#ifdef DEBUG_DIR_STRUCT
    dump_direntries(&fake_dir);
#endif
    printf("Parsed %d lines\n", lct);

    return 0;
}
