/*
 * SDPL write.c: routines to commit changes into a layer
 *
 *   Copyright (C) 2001 - 2007, Jocelyn MAYER <l_indien at magic dot fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * Write layer infos routines
 *
 */

/**SECTION: sysutils */
/**SECTION: sdpl : sysutils */
/**SECTION: editTools : sdpl */

#ifdef __KERNEL__
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/unistd.h>
#include <linux/slab.h>
#include <linux/hdreg.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/blk.h>
#include <linux/errno.h>
/* We need adler32 from zlib */
#if CONFIG_UNIZLIB
#include <linux/zlib.h>
#else
#include "adler32.h"
#endif

#else /* __KERNEL__ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
/* This is the GLIBC definition to use 64 bits file support */
#if CONFIG_BLK_SDPL_64BITS
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 1
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 1
#endif
#else /* CONFIG_BLK_SDPL_64BITS */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 0
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 0
#endif
#endif /* CONFIG_BLK_SDPL_64BITS */
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <linux/types.h>

/* We need adler32 from zlib */
#include <zlib.h>

#endif

#include <linux/version.h>
#include <linux/sdpl.h>

#ifdef __KERNEL__
void *sdpl_realloc (void *ptr, int old_size, int new_size)
{
    void *new;

    if ((new = sdpl_malloc(new_size)) == NULL)
        return NULL;
    memcpy(new, ptr, min(old_size, new_size));
    sdpl_free(ptr);
    return new;
}
#endif

#ifdef __KERNEL__

/** Helper to get the same API than the one used by the kernel
 *  at boot time
 * @param bdev file descriptor
 * @param n    sector to write
 * @param sect structure with a pointer to the buffer to write
 * @return 0, -1 if there was an error
 */
static int write_dev_sector (kdev_t dev, unsigned long n, Sector *sector)
{
    struct buffer_head *bh;
    char *datas = BUFFER(sector, n);
    char *dest;

    fsync_dev(dev);
    /* WARNING: ll_rd_wr only handle 1024 bytes bloc size */
    bh = bread(dev, n/2, 1024);
    if (!bh) {
        sdpl_info("Unable to retrieve sector %lu", n);
        return -1;
    }
    /* Increment use count */
    get_bh(bh);
    dest = bh_kmap(bh);
    memcpy(dest + ((n & 1) * 512), datas, 512);
    bh_kunmap(bh);

    mark_buffer_uptodate(bh, 1);
    mark_buffer_dirty(bh);
    brelse(bh);
    fsync_dev(dev);

    return 0;
}

#endif /* __KERNEL__ */

#if !CONFIG_BLK_SDPL_EMBEDDED || 1
/** Compare the sector currently buffered and commit writes
 *  and load the needed one, if it's not the same
 * @param sect   sector currently buffered
 * @param offset offset of the data we have to access
 * @return the sector retrieved
 */
static std_type test_sector (kdev_t bdev, Sector *sector, std_type sect,
                             std_type offset)
{
    std_type try_sect;
    static block_dev_t *bdevice;

    if ((try_sect = (offset / 512)) != sect) {
        sdpl_debug("Sector " STD_ATTRIB " to write ("
                   STD_ATTRIB " - %08lX)...\n", try_sect,
                   sect, (unsigned long)offset);
        /* Save currently edited sector */
        write_dev_sector(bdev, sect, sector);
        /* get the one we wanna patch */
        ALLOC_BLOCKDEV(bdevice, bdev);
        read_dev_sector(bdevice, try_sect, sector);
        RELEASE_BLOCKDEV(bdevice);
    }

    return try_sect;
}
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */

/** This saves the main information bloc of the layer
 * @param SDPL_map the map to save
 * @param bdev the bloc device on which save informations
 * @param sector sector currently in use
 * @param cur_sec sector number
 * @return 0, -1 in case of error
 */
static int save_SDPL_infos (struct SDPL_map *SDPL_map, kdev_t bdev,
                            Sector *sector, std_type *cur_sect)
{
    SDPL_br_t *MBR = SDPL_map->mbr;
    union SDPL_infos *infos = SDPL_map->infos;

    /* Remember that data are saved as BIG-ENDIAN */
    MBR->magic = htons(SDPL_map->magic);
    MBR->info_size = htons((int)SDPL_map->info_size);
#if !CONFIG_BLK_SDPL_EMBEDDED
    switch (SDPL_map->info_size) {
    case 16:
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        infos->infos16.bloc_size = htons((u16)SDPL_map->bloc_size);
        infos->infos16.SDPL_size = htons((u16)SDPL_map->SDPL_size);
        infos->infos16.nparts = htons((u16)NPARTS(SDPL_map));
        infos->infos16.table_offset = htons((u16)SDPL_map->pentries_offset);
        infos->infos16.bitmap_offset = htons((u16)SDPL_map->bitmap_offset);
        infos->infos16.data_offset = htons((u16)SDPL_map->data_offset);
#if !CONFIG_BLK_SDPL_EMBEDDED
        break;
    case 32:
        infos->infos32.bloc_size = htonl((u32)SDPL_map->bloc_size);
        infos->infos32.SDPL_size = htonl((u32)SDPL_map->SDPL_size);
        infos->infos32.nparts = htonl((u32)NPARTS(SDPL_map));
        infos->infos32.table_offset = htonl((u32)SDPL_map->pentries_offset);
        infos->infos32.bitmap_offset = htonl((u32)SDPL_map->bitmap_offset);
        infos->infos32.data_offset = htonl((u32)SDPL_map->data_offset);
        break;
#if CONFIG_BLK_SDPL_64BITS
    case 64:
        infos->infos64.bloc_size = htonll((u64)SDPL_map->bloc_size);
        infos->infos64.SDPL_size = htonll((u64)SDPL_map->SDPL_size);
        infos->infos64.nparts = htonll((u64)NPARTS(SDPL_map));
        infos->infos64.table_offset = htonll((u64)SDPL_map->pentries_offset);
        infos->infos64.bitmap_offset = htonll((u64)SDPL_map->bitmap_offset);
        infos->infos64.data_offset = htonll((u64)SDPL_map->data_offset);
        break;
#endif /* CONFIG_BLK_SDPL_64BITS */
    }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    write_dev_sector(bdev, 1, sector);

    return 0;
}

/** This saves all partitions informations of a layer
 * @param SDPL_map the map to save
 * @param bdev the bloc device on which save informations
 * @param sector sector currently in use
 * @param cur_sec sector number
 * @return 0, -1 in case of error
 */
static int save_parts_infos (struct SDPL_map *SDPL_map, kdev_t bdev,
                             Sector *sector, std_type *cur_sect)
{
    /* Partitions infos */
    std_type poffset = 0, bitmap_size;
    std_type i, j;
    std_type offset;
    SDPL_br_t *MBR = SDPL_map->mbr;
    union part_entry *pentry;
    unsigned char *pmap, *bitmap = NULL;
    unsigned char bmap;

    i = 0;
    bitmap_size = SDPL_map->SDPL_size / 8;
    bitmap_size = ((bitmap_size + 1023) & ~1023);
#if !CONFIG_BLK_SDPL_EMBEDDED
    if (EMBEDDED(SDPL_map) == 0) {
        poffset = 0x400 + bitmap_size +
            (NPARTS(SDPL_map) * sizeof(union part_entry));
        poffset = ((poffset + 1023) & ~1023);
    }
#endif
    offset = SDPL_map->pentries_offset;
    while (i < NPARTS(SDPL_map)) {
        /* In embedded case, offset has already been fixed */
#if !CONFIG_BLK_SDPL_EMBEDDED
        if (EMBEDDED(SDPL_map) == 0)
            SDPL_map->poffset[i] = poffset;
        /* Don't forget to get the right sector */
        if (offset < 512)
            *cur_sect = test_sector(bdev, sector, *cur_sect, offset + 512);
        else
            *cur_sect = test_sector(bdev, sector, *cur_sect, offset);
#endif
        SDPL_map->pentries[i] =
            (void *)((offset & 511) + BUFFER(sector, *cur_sect));
        pentry = SDPL_map->pentries[i];
#if !CONFIG_BLK_SDPL_EMBEDDED
        switch(SDPL_map->info_size) {
        case 16:
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
            pentry->pentry16.magic = htons(SDPL_map->pmagic[i]);
            pentry->pentry16.flags = htons((u16)SDPL_map->pflags[i]);
            pentry->pentry16.size = htons((u16)SDPL_map->psize[i]);
            pentry->pentry16.map_offset = htons((u16)SDPL_map->poffset[i]);
            pentry->pentry16.next_part = htons((u16)SDPL_map->pnext[i]);
#if !CONFIG_BLK_SDPL_EMBEDDED
            break;
        case 32:
            pentry->pentry32.magic = htons(SDPL_map->pmagic[i]);
            pentry->pentry32.flags = htons((u16)SDPL_map->pflags[i]);
            pentry->pentry32.size = htonl((u32)SDPL_map->psize[i]);
            pentry->pentry32.map_offset = htonl((u32)SDPL_map->poffset[i]);
            pentry->pentry32.next_part = htonl((u32)SDPL_map->pnext[i]);
            break;
#if CONFIG_BLK_SDPL_64BITS
        case 64:
            pentry->pentry64.magic = htons(SDPL_map->pmagic[i]);
            pentry->pentry64.flags = htons((u16)SDPL_map->pflags[i]);
            pentry->pentry64.size = htonll((u64)SDPL_map->psize[i]);
            pentry->pentry64.map_offset = htonll((u64)SDPL_map->poffset[i]);
            pentry->pentry64.next_part = htonll((u64)SDPL_map->pnext[i]);
            break;
#endif /* CONFIG_BLK_SDPL_64BITS */
        }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        pmap = (void *)((unsigned char *)MBR + SDPL_map->poffset[i]);
        offset = SDPL_map->poffset[i];
        for (j = 0; j < SDPL_map->psize[i]; j++) {
#if !CONFIG_BLK_SDPL_EMBEDDED
        if (offset < 512)
            *cur_sect = test_sector(bdev, sector, *cur_sect, offset + 512);
        else
            *cur_sect = test_sector(bdev, sector, *cur_sect, offset);
#endif
            pmap = (BUFFER(sector, *cur_sect) + (offset & 511));
#if !CONFIG_BLK_SDPL_EMBEDDED
            switch(SDPL_map->info_size) {
            case 16:
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
                 *((u16 *)pmap) = htons((u16)SDPL_map->pbloc[i][j]);
                offset += 2;
#if !CONFIG_BLK_SDPL_EMBEDDED
                break;
            case 32:
                 *((u32 *)pmap) = htonl((u32)SDPL_map->pbloc[i][j]);
                offset +=4;
                break;
#if CONFIG_BLK_SDPL_64BITS
            case 64:
                 *((u64 *)pmap) = htonll((u64)SDPL_map->pbloc[i][j]);
                offset += 8;
                break;
#endif /* CONFIG_BLK_SDPL_64BITS */
            }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        }
        poffset += SDPL_map->psize[i] * sizeof(std_type);
        offset = SDPL_map->pnext[i];
        i++;
    }
    /* Save real flags !!! */
    offset = SDPL_map->bitmap_offset;
    j = 0;
    bmap = 0;
    for (i = 0; i < SDPL_map->SDPL_size; i++) {
        if ((SDPL_map->bflags[i] & SDPL_RW) != 0) {
            bmap |= (1 << j);
        }
        j++;
        if (j == 8) {
#if !CONFIG_BLK_SDPL_EMBEDDED
        if (offset < 512)
            *cur_sect = test_sector(bdev, sector, *cur_sect, offset + 512);
        else
            *cur_sect = test_sector(bdev, sector, *cur_sect, offset);
#endif
            bitmap = BUFFER(sector, *cur_sect) + (offset & 511);
            *bitmap = bmap;
            offset++;
            bmap = 0;
            j = 0;
        }
    }
#if !CONFIG_BLK_SDPL_EMBEDDED
    if (offset < 512)
        *cur_sect = test_sector(bdev, sector, *cur_sect, offset + 512);
    else
        *cur_sect = test_sector(bdev, sector, *cur_sect, offset);
#endif
    bitmap = BUFFER(sector, *cur_sect) + (offset & 511);
    *bitmap = bmap;
    write_dev_sector(bdev, *cur_sect, sector);

    return 0;
}

#if !CONFIG_BLK_SDPL_EMBEDDED
static int realloc_partition (struct SDPL_map *SDPL_map, std_stype numparts)
{
    union part_entry **tmp_entries;
    std_type old_nparts;

    /* TODO: test NULL for each realloc */
    old_nparts = NPARTS(SDPL_map);
    /* WARNING: we suppose that we can do that (nparts don't become < 0) */
    NPARTS(SDPL_map) += numparts;
    tmp_entries =
        sdpl_realloc(SDPL_map->pentries,
                     old_nparts * sizeof(union part_entry *),
                     NPARTS(SDPL_map) * sizeof(union part_entry *));
    if (tmp_entries == NULL) {
        NPARTS(SDPL_map) -= numparts;
        return -ENOMEM;
    }
    SDPL_map->pentries = tmp_entries;
    SDPL_map->pmagic = sdpl_realloc(SDPL_map->pmagic,
                                    old_nparts * sizeof(u16),
                                    NPARTS(SDPL_map) * sizeof(u16));
    SDPL_map->pflags = sdpl_realloc(SDPL_map->pflags,
                                    old_nparts * sizeof(u16),
                                    NPARTS(SDPL_map) * sizeof(u16));
    SDPL_map->psize = sdpl_realloc(SDPL_map->psize, old_nparts *
                                   sizeof(std_type),
                                   NPARTS(SDPL_map) * sizeof(std_type));
    SDPL_map->poffset = sdpl_realloc(SDPL_map->poffset,
                                     old_nparts * sizeof(std_type),
                                     NPARTS(SDPL_map) * sizeof(std_type));
    SDPL_map->pnext = sdpl_realloc(SDPL_map->pnext,
                                   old_nparts * sizeof(std_type),
                                   NPARTS(SDPL_map) * sizeof(std_type));
    SDPL_map->pbloc = sdpl_realloc(SDPL_map->pbloc,
                                   old_nparts * sizeof(std_type *),
                                   NPARTS(SDPL_map) * sizeof(std_type *));
    return 0;
}

int add_spartition (struct SDPL_map *SDPL_map, std_type nb_blocs)
{
#if 0
    union part_entry **pentries = SDPL_map->pentries, **tmp_entries;
    std_type old_nparts;
#else
    union part_entry **pentries;
#endif
    std_type i, j;
    int ret;

#if 0
    old_nparts = NPARTS(SDPL_map);
    NPARTS(SDPL_map)++;
    tmp_entries =
        sdpl_realloc(pentries, old_nparts * sizeof(union part_entry *),
                     NPARTS(SDPL_map) * sizeof(union part_entry *));
    if (tmp_entries == NULL) {
        NPARTS(SDPL_map)--;
        return -ENOMEM;
    }
    pentries = tmp_entries;
#else
    ret = realloc_partition(SDPL_map, 1);
    if (ret < 0)
        return ret;
    pentries = SDPL_map->pentries;
#endif
    if (EMBEDDED(SDPL_map) == 1) {
        pentries[NPARTS(SDPL_map) - 1] =
            (void *)((unsigned char *)SDPL_map->mbr + 0x08 +
                     (NPARTS(SDPL_map) - 1) * 0x10);
    } else {
        /* TO FIX for non embedded cases !!! */
        std_type bitmap_size;

        bitmap_size = SDPL_map->SDPL_size / 8;
        bitmap_size = ((bitmap_size + 1023) & ~1023);
        pentries[NPARTS(SDPL_map) - 1] =
            (void *)((unsigned char *)SDPL_map->mbr + 0x400 + bitmap_size +
                     NPARTS(SDPL_map) * sizeof(union part_entry));
    }
#if 0
    SDPL_map->pmagic = sdpl_realloc(SDPL_map->pmagic,
                                    old_nparts * sizeof(u16),
                                    NPARTS(SDPL_map) * sizeof(u16));
#endif
    SDPL_map->pmagic[NPARTS(SDPL_map) - 1] = 0xE700 | SDPL_VERSION;
#if 0
    SDPL_map->pflags = sdpl_realloc(SDPL_map->pflags,
                                    old_nparts * sizeof(u16),
                                    NPARTS(SDPL_map) * sizeof(u16));
#endif
    SDPL_map->pflags[NPARTS(SDPL_map) - 1] = SDPL_RW;
#if 0
    SDPL_map->psize = sdpl_realloc(SDPL_map->psize, old_nparts *
                                   sizeof(std_type),
                                   NPARTS(SDPL_map) * sizeof(std_type));
#endif
    SDPL_map->psize[NPARTS(SDPL_map) - 1] = nb_blocs;
#if 0
    SDPL_map->poffset = sdpl_realloc(SDPL_map->poffset,
                                     old_nparts * sizeof(std_type),
                                     NPARTS(SDPL_map) * sizeof(std_type));
    SDPL_map->pnext = sdpl_realloc(SDPL_map->pnext,
                                   old_nparts * sizeof(std_type),
                                   NPARTS(SDPL_map) * sizeof(std_type));
#endif
    if (EMBEDDED(SDPL_map) == 1) {
        SDPL_map->poffset[NPARTS(SDPL_map) - 1] =
            0x48 + (NPARTS(SDPL_map) - 1) * 0x50;
        if (NPARTS(SDPL_map) < 4)
            SDPL_map->pnext[NPARTS(SDPL_map) - 1] =
                0x08 + (NPARTS(SDPL_map) * 0x10);
        else
            SDPL_map->pnext[NPARTS(SDPL_map) - 1] = 0x00;
    } else {
        /* TO FIX for non embedded cases !!! */
        std_type bitmap_size;

        bitmap_size = SDPL_map->SDPL_size / 8;
        bitmap_size = ((bitmap_size + 1023) & ~1023);
        SDPL_map->pnext[NPARTS(SDPL_map) - 1] = 0x400 + bitmap_size +
            (NPARTS(SDPL_map) * sizeof(union part_entry));
    }
#if 0
    SDPL_map->pbloc = sdpl_realloc(SDPL_map->pbloc,
                                   old_nparts * sizeof(std_type *),
                                   NPARTS(SDPL_map) * sizeof(std_type *));
#endif
    if (nb_blocs > 0) {
        /* Allocate some of the free blocs for partition */
        SDPL_map->pbloc[NPARTS(SDPL_map) - 1] =
            sdpl_malloc(nb_blocs * sizeof(std_type));
        j = 0;
        i = 0;
        while (i < SDPL_map->SDPL_size && j < nb_blocs) {
            if (SDPL_map->bflags[i] == EMPTY_BLOC) {
                SDPL_map->bflags[i] =
                    SDPL_map->pflags[NPARTS(SDPL_map) - 1] & SDPL_PUBLIC;
                SDPL_map->pbloc[NPARTS(SDPL_map) - 1][j] = i;
                i++;
            } else {
                i++;
                continue;
            }
            j++;
        }
    } else {
        SDPL_map->pbloc[NPARTS(SDPL_map) - 1] = NULL;
    }
    return 0;
}

int remove_spartition (struct SDPL_map *SDPL_map, std_type part_num)
{
    std_type real_bloc, i;

    /* Mark all blocs as free */
    for (i = 0; i < SDPL_map->psize[part_num]; i++) {
        real_bloc = SDPL_map->pbloc[part_num][i];
        SDPL_map->bflags[real_bloc] = EMPTY_BLOC;
    }
    /* remove partition from partition list */
    return realloc_partition(SDPL_map, -1);
}
#endif

int remove_bloc_from_part (struct SDPL_map *SDPL_map, std_type part_num,
                           std_type num_bloc)
{
    std_type real_bloc, old_size, i, j;
    std_type *new_pbloc, *old_pbloc;

    if (num_bloc > SDPL_map->psize[part_num]) {
        sdpl_error("Bloc to free " STD_ATTRIB " out of partition "
                   STD_ATTRIB "\n", num_bloc, SDPL_map->psize[part_num]);
        return -1;
    }

    real_bloc = SDPL_map->pbloc[part_num][num_bloc];
    old_size = SDPL_map->psize[part_num];
    SDPL_map->psize[part_num]--;
    old_pbloc = SDPL_map->pbloc[part_num];
    new_pbloc = sdpl_malloc(SDPL_map->psize[part_num] * sizeof(std_type));
    for (i = 0, j = 0; i < old_size; i++) {
        if (i != num_bloc) {
            new_pbloc[j] = old_pbloc[i];
            j++;
        } else {
                sdpl_debug("Skipping bloc " STD_ATTRIB
                           " (" STD_ATTRIB ")\n", i, real_bloc);
        }
    }
    SDPL_map->pbloc[part_num] = new_pbloc;
    SDPL_map->bflags[real_bloc] = EMPTY_BLOC;
    sdpl_free(old_pbloc);
    return 0;
}

int add_blocs_to_part (struct SDPL_map *SDPL_map, std_type part_num,
                       std_type nb_blocs)
{
    std_type i, j;
    std_type old_size, new_size, max_size;
    std_type *tmp_pbloc /*, *old_pbloc */;

    max_size = count_free_blocs(SDPL_map);
#if 0
#if !CONFIG_BLK_SDPL_EMBEDDED
    if (EMBEDDED(SDPL_map) == 1)
#endif
        max_size = min((std_type)40, max_size);
#endif

    if (nb_blocs > max_size) {
        sdpl_error("resize of part " STD_ATTRIB ": " STD_ATTRIB
                   " > max size: " STD_ATTRIB "\n", part_num, nb_blocs,
                   max_size);
        return -EINVAL;
    }
    /* Allocate some of the free blocs for partition */
    old_size = SDPL_map->psize[part_num];
    new_size = SDPL_map->psize[part_num] + nb_blocs;
    tmp_pbloc = sdpl_realloc(SDPL_map->pbloc[part_num],
                             old_size * sizeof(std_type),
                             new_size * sizeof(std_type));
    if (tmp_pbloc == NULL) {
        sdpl_error("Unable to allocate new SDPL map\n");
        return -ENOMEM;
    }

    SDPL_map->psize[part_num] = new_size;
    SDPL_map->pbloc[part_num] = tmp_pbloc;

    i = 0;
    j = old_size;
    /*
     * The first test is not needed,
     * as we already tested that we have some free blocs to allocate
     */
    while (/* i < SDPL_map->SDPL_size && */ j < new_size) {
        if (SDPL_map->bflags[i] == EMPTY_BLOC) {
            sdpl_debug("Bloc " STD_ATTRIB " is empty (" STD_ATTRIB
                       ").\n", i, j);
            SDPL_map->bflags[i] = SDPL_map->pflags[part_num] & SDPL_PUBLIC;
            SDPL_map->pbloc[part_num][j] = i;
            j++;
        }
        i++;
    }
    return 0;
}

void mark_partition_bootable (struct SDPL_map *SDPL_map, std_type part_num)
{
    std_type i;

    for (i = 0; i < NPARTS(SDPL_map); i++) {
        if (i == part_num) {
#ifdef __KERNEL__
            sdpl_info("Partition " STD_ATTRIB " is set bootable\n", i);
#endif
            SDPL_map->pflags[i] |= SDPL_BOOT;
        } else {
            SDPL_map->pflags[i] &= ~SDPL_BOOT;
        }
    }
}

void toggle_flag (u16 *flag, int toggle)
{
    if (toggle == -1) {
        *flag = ((*flag & SDPL_RW) == SDPL_RW ? SDPL_RO : SDPL_RW);
    } else {
        *flag = (toggle == 0 ? SDPL_RO : SDPL_RW);
    }
}

void toggle_part_flag (struct SDPL_map *SDPL_map, std_type part_num,
                       int toggle)
{
    std_type i;
    char flag;

    flag = SDPL_map->pflags[part_num] & (SDPL_PRIVATE | SDPL_BOOT);
    toggle_flag(&SDPL_map->pflags[part_num], toggle);
    SDPL_map->pflags[part_num] |= flag;
    sdpl_info("Set " STD_ATTRIB " bloc flags to %d\n",
              SDPL_map->psize[part_num], flag);
    for (i = 0; i < SDPL_map->psize[part_num]; i++)
        SDPL_map->bflags[SDPL_map->pbloc[part_num][i]] =
                SDPL_map->pflags[part_num] & SDPL_PUBLIC;
}

void save_sdpl (struct SDPL_map *SDPL_map, kdev_t bdev,
                Sector *sector, std_type *cur_sect)
{
    char tmp_buffer[512];
    SDPL_br_t *MBR = SDPL_map->mbr;
    block_dev_t *_bdev;

    sdpl_info("Going to save layer\n");
    ALLOC_BLOCKDEV(_bdev, bdev);
    *cur_sect = test_sector(bdev, sector, *cur_sect, 512);
    SDPL_map->magic = 0xE700 | SDPL_VERSION;
    save_SDPL_infos(SDPL_map, bdev, sector, cur_sect);
    save_parts_infos (SDPL_map, bdev, sector, cur_sect);
#if !CONFIG_BLK_SDPL_EMBEDDED
    *cur_sect = test_sector(bdev, sector, *cur_sect, 512);
#endif
    memcpy(tmp_buffer, BUFFER(sector, 1), 512);
    /* Recalculate CRC */
    SDPL_map->CRC = (unsigned int)adler32(1, tmp_buffer + 8, 512 - 8);
    MBR->CRC = htonl(SDPL_map->CRC);
    memcpy(BUFFER(sector, 1) + 4, &MBR->CRC, 4);
    memcpy(tmp_buffer + 4, &MBR->CRC, 4);
    /* Save duplicate infos */
    *cur_sect = test_sector(bdev, sector, *cur_sect, 0);
    memcpy(BUFFER(sector, 0), tmp_buffer, 512);
    write_dev_sector(bdev, 0, sector);
    RELEASE_BLOCKDEV(_bdev);
    fsync_dev(bdev);
}

/* 64 bits endianness conversion */
#if CONFIG_BLK_SDPL_64BITS
/** Converts a 64 bit data from network to host endianness
 */
u64 ntohll (u64 data)
{
    unsigned char *p = (unsigned char *)(&data);
    u64 p1;
    u32 p2;

    p1 = ntohl(*((u32 *)p));
    p2 = ntohl(*((u32 *)(p + 4)));

    return (p1 << 32) + p2;
}

/** Converts a 64 bit data from host to network endianness
 */
u64 htonll (u64 data)
{
    unsigned char *p = (unsigned char *)(&data);
    u64 p1;
    u32 p2;

    p1 = htonl(*((u32 *)p));
    p2 = htonl(*((u32 *)(p + 4)));

    return (p1 << 32) + p2;
}
#endif /* CONFIG_BLK_SDPL_64BITS */

/** Count free blocs in a layer
 */
std_type count_free_blocs (struct SDPL_map *SDPL_map)
{
    std_type count = 0, bloc;

    /* Trying to be safe, while in kernel space... */
    if (SDPL_map->bflags != NULL) {
        for (bloc = 0; bloc < SDPL_map->SDPL_size; bloc++) {
            if (SDPL_map->bflags[bloc] == EMPTY_BLOC)
                count++;
        }
    }
    return count;
}

#if !CONFIG_BLK_SDPL_EMBEDDED && !defined(__KERNEL__)
void fix_bflags (struct SDPL_map *SDPL_map)
{
    sdpl_free(SDPL_map->bflags);
    SDPL_map->bflags = sdpl_malloc(SDPL_map->SDPL_size * sizeof(u16));
    memset (SDPL_map->bflags, EMPTY_FILL, SDPL_map->SDPL_size * sizeof(u16));
}
#endif /* !CONFIG_BLK_SDPL_EMBEDDED && !defined(__KERNEL__) */

#if !CONFIG_BLK_SDPL_EMBEDDED && !defined(__KERNEL__)
void fix_SDPL_infos (struct SDPL_map *SDPL_map)
{
    SDPL_map->magic = 0xE700 | SDPL_VERSION;
    SDPL_map->CRC = 0;
    if (EMBEDDED(SDPL_map)) {
        SDPL_map->SDPL_size = 64;
        SDPL_map->bloc_size = (SDPL_map->st_size) /
            (SDPL_map->SDPL_size * 512);
    } else {
        SDPL_map->bloc_size = 2;
    }
    SDPL_map->SDPL_size = (SDPL_map->st_size) / (SDPL_map->bloc_size * 512);
    /* I keep this, in case we include the routine in embedded version */
#if !CONFIG_BLK_SDPL_EMBEDDED
    NPARTS(SDPL_map) = 0;
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    fix_bflags(SDPL_map);
}
#endif /* !CONFIG_BLK_SDPL_EMBEDDED && !defined(__KERNEL__) */

/* Find best bloc size for a layer, given layer size */
#if !CONFIG_BLK_SDPL_EMBEDDED && !defined(__KERNEL__)
unsigned long get_ressize (unsigned long tot_size, int size, int info_size,
                           int embedded, std_type *pe_size,
                           std_type *pm_size, std_type *bm_size)
{
    std_type nparts, pentry_size, pmap_size, bitmap_size;

    if (embedded == 1)
        return 1024;

        /* Let's consider we can have 1 partition / Mbyte */
    nparts = max(4, (tot_size) >> 20);
    pentry_size = nparts * (info_size / 8);
    pentry_size = ((pentry_size + 1023) & ~1023);
    if (pe_size != NULL)
        *pe_size = pentry_size;
    /* Let's consider we need one mapping bloc / sector */
    pmap_size = (tot_size / 1024 / 1024 / 512) *
        sizeof(union part_entry) * nparts;
    pmap_size = ((pmap_size + 1023) & ~1023);
    if (pm_size != NULL)
        *pm_size = pmap_size;
    bitmap_size = (tot_size >> 10) / 8;
    bitmap_size = ((bitmap_size + 1023) & ~1023);
    if (bm_size != NULL)
        *bm_size = bitmap_size;

    return 1024 + pentry_size + pmap_size + bitmap_size;
}

int get_best_size (const unsigned long tot_size, unsigned long *reserved,
                   const int min_blocs, const int max_blocs, int *nblocs,
                   int info_size, int embedded)
{
    unsigned long layer_size = tot_size, lost = tot_size, temp;
    int size = max_blocs, best = max_blocs, n;

    /* Have to adapt reserved size to constraints... */
    while (size >= min_blocs) {
        *reserved = get_ressize(tot_size, size, info_size, embedded,
                                NULL, NULL, NULL);
        layer_size = tot_size - *reserved;
        n = layer_size / (size * 512);
        temp = layer_size - (size * n * 512);
        if (temp < lost) {
            best = size;
            lost = temp;
        }
        layer_size = tot_size - *reserved;
        size--;
    }
    if (nblocs != NULL)
        *nblocs = layer_size / (best * 512);

    return best;
}
#endif /* !CONFIG_BLK_SDPL_EMBEDDED && !defined(__KERNEL__) */
