/*
 * SDPL sdplbuild.c: Utility to build a SDPL layer from a set of files
 *                   into another file
 *
 *   Copyright (C) 2001, 2002, 2003, 2004, Jocelyn MAYER <l_indien@magic.fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * Layer build tool
 *
 */

/* 
 * Usage: sdplbuild layer_name description_file
 *
 * Description file format:
 * Everything after a '#' is ignored
 *
 * # To define the layer file name:
 * layer layername
 * # Option to build an embedded layer
 * embedded
 * # Size for the layer (in kBytes):
 * lsize 8192
 * # The parameters minspb, maxspb, minblocs, maxblocs are optional
 * # Bloc size min (in sectors):
 * minspb 40
 * # Bloc size max:
 * maxspb 64
 * # Number (min) of blocs:
 * minblocs 2
 * # Number (max) of blocs:
 * maxblocs 128
 * # The pattern to use to pad empty parts of the layer:
 * pad 88
 * # Files to insert in the layer
 * file file1
 * file file2
 * file file3
 *
 *
 * To build an embedded layer,
 * just put 'e' or 'E' before the layer size
 *
 */

/* 
 * Do we want to map input files ?
 * Note that this won't work for big files (> 1Go)
 */
//#define USE_MMAP

/* Needed for getline: */
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef USE_MMAP
#include <sys/mman.h>
#endif

#include <linux/types.h>
#include <linux/version.h>

#include <linux/sdpl.h>

#if 0
#undef sdpl_debug
#define sdpl_debug(fmt, args...) { fprintf(stderr, fmt , ##args); } while (0)
#endif

/* Internal file description */
typedef struct file_desc_t {
    int fd;
    char *name;
    unsigned long real_size;
    int size;
    int spb;
    struct file_desc_t *next;
} file_desc_t;

static file_desc_t *file_list = NULL, *flist_end = NULL;
static int embedded = 0;
static char pattern;

#if 0
void dump_files (void)
{
    file_desc_t *c;
    for (c = file_list; c != NULL; c = c->next)
        sdpl_debug("file: %d %s\n", c->fd, c->name);
}
#endif

/* Parser definition */
static int min_spb = -1, max_spb = -1, min_blocs = -1, max_blocs = -1;
static unsigned long file_total;

typedef struct {
    const char *name;
    int exact_match;
    int (*command)(char *line);
} command_t;

static int inqueue_file (const char *name);

static int get_layer_name (char *line);
static int get_embed_flag (char *line);
static int get_layer_size (char *line);
static int get_min_spb (char *line);
static int get_max_spb (char *line);
static int get_spb (char *line);
static int get_min_blocs (char *line);
static int get_max_blocs (char *line);
static int get_padding_char (char *line);
static int get_filename (char *line);
static int skip_comments (char *line);

static command_t commands[] = {
    /* Please, keep the comment mark as first one */
    { "#", 1, skip_comments },
    { "layer", 1, get_layer_name },
    { "embedded", 0, get_embed_flag },
    { "lsize", 1, get_layer_size },
    { "minspb", 1, get_min_spb },
    { "maxspb", 1, get_max_spb },
    { "spb", 1, get_spb },
    { "minblocs", 1, get_min_blocs },
    { "maxblocs", 1, get_max_blocs },
    { "pad", 1, get_padding_char },
    { "file", 1, get_filename },
    { NULL, 0, NULL },
};

static int get_num (const char *line)
{
    const char *pos = line;
    
    while (isspace(*pos))
        pos++;
    if (isdigit(*pos))
        return atoi(pos);

    return -1;
}

static int get_string (char *line, char **nline)
{
    char *pos = line, *end;

    while (isspace(*pos))
        pos++;
    end = pos;
    /* commands[0] is supposed to be comment mark */
    while (!isspace(*end) && 
           strstr(commands[0].name, end) != commands[0].name)
        end++;
    *end = '\0';
    if (nline != NULL)
        *nline = pos;

    return strlen(pos);
}       

static int get_layer_name (char *line)
{
    char *sname;

    get_string(line, &sname);
    file_list->name = strdup(sname);
    file_list->size = -1;
    file_list->spb = 0;

    return 0;
}

static int get_embed_flag (char *line)
{
    embedded = 1;
    if (min_spb == -1)
        min_spb = 47;
    if (max_spb == -1)
        max_spb = 64;

    return 0;
}

static int get_layer_size (char *line)
{
    unsigned long size, reserved = 1024;
    int sizetype = sizeof(std_type) * 8;

    if (strcmp(file_list->name, "") == 0) {
        fprintf(stderr, "Should define layer name first\n");
        return -1;
    }
    if ((size = get_num(line)) == (-1UL)) {
        fprintf(stderr, "Attempt to get a number here: %s\n", line);
        return -1;
    }
    size *= 1024;
    if (size < 2048) {
        fprintf(stderr, "Layer is too small: %lu bytes\n", size);
        return -1;
    }
    file_list->real_size = size;
    if (file_list->spb == 0) {
        if (embedded == 1) {
            if (min_spb == -1)
                min_spb = 47;
            if (max_spb == -1)
                max_spb = 64;
            sizetype = 16;
        } else {
            if (max_blocs != -1 && max_blocs != 0) {
                max_spb = (size / max_blocs) >> 9;
            }
            if (min_blocs != -1 && min_blocs != 0) {
                min_spb = (size / min_blocs) >> 9;
            }
//        if (spb_min == -1 || spb_max == -1) {
            fprintf(stderr, "Sorry, implementation not complete:\n"
                    "Can only generate embedded layers\n");
            return -1;
//        }
        }
        file_list->size = get_best_size(file_list->real_size, &reserved,
                                        min_spb, max_spb, &file_list->spb, 
                                        sizetype, embedded);
    } else {
        std_type blocs = (file_list->real_size - 1024) / 
            (file_list->spb * 512);
        file_list->size = blocs;
        reserved = file_list->real_size - 
            (file_list->size * file_list->spb * 512) - 1024;
    }

    fprintf(stderr, "Layer %s - real size %lu - spb %d - size %d"
            " - lost %lu\n",
            file_list->name, file_list->real_size, file_list->spb,
            file_list->size, file_list->real_size - 
            (file_list->size * file_list->spb * 512));

    return size;
}

static int get_min_spb (char *line)
{
    if ((min_spb = get_num(line)) < 0) {
        fprintf(stderr, "Attempt to get a number here: %s\n", line);
        return -1;
    }
    return min_spb;
}

static int get_max_spb (char *line)
{
    if ((max_spb = get_num(line)) < 0) {
        fprintf(stderr, "Attempt to get a number here: %s\n", line);
        return -1;
    }
    return max_spb;
}

static int get_spb (char *line)
{
    if ((file_list->spb = get_num(line)) < 0) {
        fprintf(stderr, "Attempt to get a number here: %s\n", line);
        return -1;
    }
    return file_list->spb;
}

static int get_min_blocs (char *line)
{
    if ((min_blocs = get_num(line)) < 0) {
        fprintf(stderr, "Attempt to get a number here: %s\n", line);
        return -1;
    }
    return min_blocs;
}

static int get_max_blocs (char *line)
{
    if ((max_blocs = get_num(line)) < 0) {
        fprintf(stderr, "Attempt to get a number here: %s\n", line);
        return -1;
    }
    return max_blocs;
}

static int get_padding_char (char *line)
{
    char *pos;
    int pad;
    
    if ((pad = get_num(line)) < 0) {
        pos = line;
        while (isspace(*pos))
            pos++;
        pattern = *pos;
    } else {
        if (pad > 255)
            pattern = 255;
        else
            pattern = pad;
    }

    return pad;
}

static int get_filename (char *line)
{
    struct stat f_stat;
    unsigned long size;
    char *sname;

    get_string(line, &sname);
    inqueue_file(sname);

    if ((flist_end->fd = open(flist_end->name, O_RDONLY)) < 0) {
        fprintf(stderr, "Can't open file: %s\n", flist_end->name);
        return -1;
    }
    if (stat(flist_end->name, &f_stat) < 0) {
        fprintf(stderr, "Unable to get file infos: %s\n",
                flist_end->name);
        return -1;
    }
    size = ((f_stat.st_size + (file_list->spb * 512) - 1) / 
            (file_list->spb * 512));
    fprintf(stderr, "Got file: %s - size: %lu -> %lu blocs "
            "(%lu)\n", flist_end->name, f_stat.st_size, size,
            size * file_list->spb * 512);
    flist_end->spb = file_list->spb;
    flist_end->real_size = f_stat.st_size;
    flist_end->size = size;
    file_total += size * file_list->spb * 512;

    return size;
}

static int skip_comments (char *line)
{
    return 0;
}

/* Always inqueue at the end of the list */
static int inqueue_file (const char *name)
{
    file_desc_t *new = malloc(sizeof(file_desc_t));

    if (new == NULL)
        return -1;

    sdpl_debug("New file: %s\n", name);
    memset(new, 0, sizeof(file_desc_t));
    new->fd = -1;
    if (name != NULL)
        new->name = strdup(name);
    if (flist_end == NULL) {
        file_list = new;
    } else {
        flist_end->next = new;
    }
    flist_end = new;

    return 0;
}

static void freeall (void)
{
    file_desc_t *current, *next;

    for (current = file_list; current != NULL; current = next) {
        sdpl_debug("Going to free fd=%d name=%s\n",
                   current->fd, current->name);
        if (current->fd > 0)
            close(current->fd);
        free(current->name);
        next = current->next;
        free(current);
    }
}

static int parse_file (FILE *stream)
{
    char *line, *pos, *end;
    ssize_t length;
    int count = 0, ret;
    command_t *command;
    int got_it;

    /* Reserve a file in structures for the layer */
    inqueue_file(NULL);

    line = NULL;
    length = 0;
    while ((ret = getline(&line, &length, stream)) != -1) {
        pos = line;
        while (isspace(*pos))
            pos++;
        got_it = 0;
        end = pos;
        /* The last condition should never be reached, but who knows... */
        while (!isspace(*end) && *end != '\0' && (end - pos) < length)
            end++;
        *end = '\0';
        /* Skip line if empty */
        if (end == pos)
            continue;
        command = commands;
        while (command->name != NULL) {
            if (command->exact_match == 1 && 
                strcasecmp(command->name, pos) == 0) {
                got_it = 1;
                break;
            }
            if (command->exact_match == 0) {
                char *pos1 = pos;
                while (pos1 < end) {
                    *pos1 = tolower(*pos1);
                    pos1++;
                }
                if (strstr(command->name, pos) == command->name) {
                    got_it = 1;
                    break;
                }
            }
            command++;
        }
        if (got_it == 0) {
            fprintf(stderr, "Unknown command: %s\n", pos);
            goto parse_error;
        }
        if (command->command(pos + strlen(command->name) + 1) < 0) {
            fprintf(stderr, "Error while executing: %s\n", command->name);
            goto parse_error;
        }
        count++;
    }
    free(line);

    if (count < 2) {
        fprintf(stderr, "Too few informations in description\n");
        goto parse_error;
    }

    if (file_total > file_list->real_size) {
        fprintf(stderr, "Files are bigger than the whole layer ! (%lu - %lu)"
                "\n", file_total, file_list->real_size);
        goto parse_error;
    }
    fprintf(stderr, "%lu bytes allocated over %lu in the layer\n", 
            file_total, file_list->real_size);

    return 0;

 parse_error:
    freeall();
    return -1;
}

static int build_SDPL_map (int fd, std_type dev_size, int numparts, 
                           int res_size)
{
    struct block_device bdev;
    Sector sect;
    /* main structure */
    struct SDPL_map *SDPL_map = NULL;
    SDPL_br_t *MBR;
    file_desc_t *current = file_list->next;
    int part = 0;
    std_type cur_sect = 0;
    int ret;

    ret = -1;
    bdev.fd = fd;
    if ((sect.sector = malloc(1024)) == NULL)
        goto build_error;

    if ((SDPL_map = malloc(sizeof(struct SDPL_map))) == NULL)
        goto build_error;
    memset(SDPL_map, 0, sizeof(struct SDPL_map));

    MBR = (SDPL_br_t *)sect.sector;
    SDPL_map->mbr = MBR;

    EMBEDDED(SDPL_map) = embedded;
    NPARTS(SDPL_map) = 0;

    SDPL_map->pentries = NULL;
#if 0 // pentries will be re-allocated when each time we add a partition...
    SDPL_map->pentries =
        sdpl_malloc(NPARTS(SDPL_map) * sizeof(union part_entry *));
    if (SDPL_map->pentries == NULL)
        goto build_error;
    memset(SDPL_map->pentries, 0,
           NPARTS(SDPL_map) * sizeof(union part_entry *));
#endif
    SDPL_map->infos = (void *)MBR->SDPL_infos;

    SDPL_map->magic = 0xE700 | SDPL_VERSION;
    SDPL_map->CRC = 0;
    SDPL_map->res_size = res_size;
    SDPL_map->data_offset = SDPL_map->res_size;
    SDPL_map->device_size = dev_size;
    SDPL_map->st_size = dev_size - SDPL_map->res_size;

    /* 
     * For embedded system, parameters are set so all information
     * can go in the boot-record
     */
    SDPL_map->SDPL_size = file_list->size;
    SDPL_map->bloc_size = file_list->spb;
    if (EMBEDDED(SDPL_map) == 1) {
        SDPL_map->info_size = 16;
        SDPL_map->pentries_offset = 0x08;
        SDPL_map->bitmap_offset = 0x190;
        numparts = 4;
    } else {
        std_type bitmap_size;

        SDPL_map->bitmap_offset = 0x400;
        bitmap_size = SDPL_map->SDPL_size / 8;
        bitmap_size = ((bitmap_size + 1023) / 1024) * 1024;
        SDPL_map->pentries_offset = 0x400 + bitmap_size;
        if (SDPL_map->info_size == 0)
            SDPL_map->info_size = sizeof(std_type) * 8;
    }
    if ((SDPL_map->bflags = 
         malloc(SDPL_map->SDPL_size * sizeof(__u16))) == NULL) 
        goto build_error;
    memset (SDPL_map->bflags, EMPTY_FILL, SDPL_map->SDPL_size * sizeof(__u16));
    
    while (current != NULL) {
        sdpl_debug("Add partition: %s (%d)\n", current->name, part);
        add_spartition(SDPL_map, current->size);
        current = current->next;
        part++;
    }
    if (EMBEDDED(SDPL_map) == 1) {
        for (; part < numparts; part++) {
            add_spartition(SDPL_map, 0);
        }
    }
    mark_partition_bootable (SDPL_map, 0);
    save_sdpl(SDPL_map, &bdev, &sect, &cur_sect);

    ret = 0;

 build_error:
    free(SDPL_map->bflags);
    free(SDPL_map->pentries);
    free(SDPL_map);
    free(sect.sector);

    return ret;
}

#define BUFLEN 65536
static int build_layer (int copy_files)
{
    u64 total = 0, max, len, copied, reserved;
    file_desc_t *current = file_list->next;
    char *buffer, *write_pos;
#ifdef USE_MMAP
    char *filemap, *pos;
#endif
    int numfiles = 0, rlen;
    u64 max_std;
    int size = sizeof(std_type);

    /* Determine the max offset we can store in std_type */
    if (embedded == 1)
	size = 2;
    switch (size) {
    case 2:
	max_std = (1 << 15) - 1;
	break;
    case 4:
	max_std = ((u64)1 << 31) - 1;
	break;
    case 8:
    default:
	max_std = ((u64)1 << 63) - 1;
	break;
    }
    max_std = (max_std << 1) + 1;

    if (!embedded) {
        fprintf(stderr, "For now, I'll just make embedded layers... Sorry\n");
        goto build_error;
    }
    /* First, allocate a BIG buffer */
    if ((buffer = malloc(BUFLEN)) == NULL) {
        fprintf(stderr, "Can't get memory\n");
        goto build_error;
    }
    memset(buffer, pattern, BUFLEN);

    /* Then, get the layer file */
    unlink(file_list->name);
    if ((file_list->fd = open(file_list->name, O_RDWR | O_EXCL | O_CREAT | 
                              O_TRUNC | O_NOFOLLOW)) < 0) {
        fprintf(stderr, "Unable to create a new layer file %s\n",
                file_list->name);
        goto build_error;
    }
    if (fchmod(file_list->fd, 0644) < 0) {
        fprintf(stderr, "can't set permissions for file %s\n", 
                file_list->name);
        goto build_error;
    }

    /* Now, we got to fill it with PAD */
    max = file_list->size * file_list->spb * 512;

    reserved = (file_list->real_size - max);
    if (reserved > max_std) {
	fprintf(stderr, "WARNING: wanted data offset > capacity of std_type\n"
		"Remapping %llu to 0x2000\n", reserved);
	reserved = 0x2000;
    }
    len = BUFLEN;
    fprintf(stderr, "Reserved %llu bytes (%d blocs - %d s/b - %ld bytes)\n", 
            reserved, file_list->size, file_list->spb, file_list->real_size);
    /* First pad reserved blocs - with ZEROES */
    memset(buffer, 0, BUFLEN);
    while (total < reserved) {
        if ((reserved - total) < BUFLEN)
            len = reserved - total;
        if (write(file_list->fd, buffer, len) < 0) {
            fprintf(stderr, "Error while preparing reserved blocs\n");
            goto build_error;
        }
        total += len;
    }
    /* Then copy files */
    if (copy_files != 0) {
        while (current != NULL) {
            numfiles++;
            copied = 0;
#ifdef USE_MMAP
            if ((filemap = mmap(NULL, current->real_size, 
                                PROT_READ, MAP_PRIVATE,
                                current->fd, 0)) == MAP_FAILED) {
                fprintf(stderr, "Can't map file: %s\n", current->name);
                goto build_error;
            }
            pos = filemap;
#endif
            while (copied < current->real_size) {
                len = BUFLEN;
                if ((current->real_size - copied) < BUFLEN)
                    len = current->real_size - copied;
#ifdef USE_MMAP
                if (filemap == MAP_FAILED) 
#endif
                    {
                        rlen = read(current->fd, buffer, len);
                        if (rlen < 0) {
                            fprintf(stderr, "Error while reading %s\n", 
                                    current->name);
                            goto build_error;
                        }
                        write_pos = buffer;
#ifdef USE_MMAP
                    } else {
                        write_pos = pos;
#endif
                    }
                if (write(file_list->fd, write_pos, rlen) < 0) {
                    fprintf(stderr, "Error while copying %s\n", current->name);
                    goto build_error;
                }
#ifdef USE_MMAP
                pos += rlen;
#endif
                total += rlen;
                copied += rlen;
            }
#ifdef USE_MMAP
            if (filemap != MAP_FAILED)
                munmap(filemap, current->real_size);
#endif
            /* Don't forget to pad it */
            memset(buffer, pattern, BUFLEN);
            max = current->spb * current->size * 512;
            len = BUFLEN;
            while (copied < max) {
                if ((max - copied) < BUFLEN)
                    len = max - copied;
                if (write(file_list->fd, buffer, len) < 0) {
                    fprintf(stderr, "Error while padding %s\n", current->name);
                    goto build_error;
                }
                total += len;
                copied += len;
            }
            current = current->next;
        }
        /* And now, pad the layer */
        max = file_list->real_size;
        len = BUFLEN;
        memset(buffer, pattern, BUFLEN);
        while (total < max) {
            if ((max - total) < BUFLEN)
                len = max - total;
            if (write(file_list->fd, buffer, len) < 0) {
                fprintf(stderr, "Error while padding %s\n", file_list->name);
                goto build_error;
            }
            total += len;
        }
    }
    build_SDPL_map(file_list->fd, file_list->real_size, numfiles,
		   reserved);

    return 0;

 build_error:
    freeall();
    return -1;
}

static void usage (const char *name)
{
    fprintf(stderr, "%s [-b] desc_file\n", name);
    fprintf(stderr, "  desc_file: name of the file describing\n");
    fprintf(stderr, "             the layer to be built\n");
    fprintf(stderr, "\t-b: only generate the boot bloc (+ pad)\n");
}

int main (int argc, char **argv)
{
    const char *filename;
    FILE *stream;
    int copy_files = 1;
    int c;
    
    while ((c =getopt(argc, argv, "b")) != -1) {
        switch (c) {
        case 'b':
            copy_files = 0;
            break;
        case 'h':
        case '?':
            usage(argv[0]);
            exit(0);
            break;
        default:
            fprintf(stderr, "Unknown option: %c\n", c);
            usage(argv[0]);
            exit(1);
            break;
        }
    }
    
    if (optind != (argc - 1)) {
        fprintf(stderr, "Bad number of arguments\n");
        usage(argv[0]);
        exit(1);
    }

    filename = argv[optind];
    if ((stream = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Wrong file name: %s\n", filename);
        exit(1);
    }
    if (parse_file(stream) < 0) {
        fprintf(stderr, "Bad description file: %s\n", filename);
        exit(1);
    }
    fclose(stream);

    if (build_layer(copy_files) < 0) {
        fprintf(stderr, "Error while building layer\n");
        exit(1);
    }
    freeall();

    exit(0);
}
