/*
 * SDPL sdpledit.c: a edit tool for sdpl layers
 *
 *   Copyright (C) 2001, 2002, 2003, 2004, Jocelyn MAYER <l_indien@magic.fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * user-mode edit tool for SDPL
 *
 * Here's the edit tool for SDPL.
 *
 */

/**SECTION: sysutils */
/**SECTION: sdpl */
/**SECTION: edit tool */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
/* This is the GLIBC definition to use 64 bits file support */
#if CONFIG_BLK_SDPL_64BITS
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 1
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 1
#endif
#else /* CONFIG_BLK_SDPL_64BITS */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 0
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 0
#endif
#endif /* CONFIG_BLK_SDPL_64BITS */
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <linux/types.h>
#include <linux/version.h>

/* We need adler32 from zlib */
#include <zlib.h>

#include <linux/sdpl.h>

/* I restrain (!) my command line to 256 characters */
#define MAX_CMD 256
/* Yet another MAGIC value (with no meaning !) */
#define DO_EXIT 211

/* Command structure definition */
struct commands {
    const char *cmd;
    const char *comments;
    int arg_min, arg_max;
    std_stype (*exec)(const char *bufargs);
};

static struct SDPL_map *SDPL_map = NULL;
static Sector sector;
static std_type cur_sect = 0;
static struct block_device *bdev;

static inline std_stype do_nothing (const char *bufargs);
static std_stype do_exit (const char *bufargs);
static std_stype do_cmd_error (const char *bufargs);
static std_stype do_print_sdpl_infos (const char *bufargs);
static std_stype do_print_sdpl_part (const char *bufargs);
static std_stype do_print_free_blocs (const char *bufargs);
static std_stype init_sdpl (const char *bufargs);
static std_stype change_info_size (const char *bufargs);
static std_stype change_bloc_size (const char *bufargs);
static std_stype change_layer_size (const char *bufargs);
#if 0
static std_stype change_part_offset (const char *bufargs);
#endif
static std_stype do_add_partition (const char *bufargs);
static std_stype do_remove_partition (const char *bufargs);
static std_stype do_resize_partition (const char *bufargs);
static std_stype do_mark_partition_bootable (const char *bufargs);
static std_stype do_toggle_part_flag (const char *bufargs);
static std_stype do_toggle_bloc_flag (const char *bufargs);
static std_stype do_save_sdpl (const char *bufargs);

/** Implemented commands... See sdpledit.h
 */
static struct commands commands[] = {
    { "h?", "Print this help", 0, 0, do_cmd_error },
    { "p", "print layer", 0, 0, do_print_sdpl_infos },
    { "d", "dump partition", 0, 1, do_print_sdpl_part },
    { "f", "dump free blocs", 0, 0, do_print_free_blocs },
    { "i", "Init partition layer", 0, 1, init_sdpl },
    { "s", "Change info size", 1, 1, change_info_size },
    { "b", "Change bloc size", 1, 1, change_bloc_size },
    { "l", "Change layer size", 1, 1, change_layer_size },
#if 0
    { "o", "Change partition infos offset", 1, 1, change_part_offset },
#endif
    { "n", "Add partition", 0, 1, do_add_partition },
    { "k", "Remove partition", 0, 1, do_remove_partition },
    { "r", "Resize partition", 1, 2, do_resize_partition },
    { "m", "Mark partition bootable", 1, 1, do_mark_partition_bootable },
    { "t", "Toggle partition flag", 1, 2, do_toggle_part_flag },
    { "y", "Toggle bloc flag", 1, 2, do_toggle_bloc_flag },
    { "w", "Save layer", 0, 0, do_save_sdpl },
    { "xq", "Exit without saving", 0, 0, do_exit },
    { NULL , NULL, 0, 0, NULL },
};

/** Helper to find the first non blank char in a buffer
 * @param buffer
 * @param pos first non blank character, or NULL
 */
static int find_not_space (const char *buffer, char **pos)
{
    char *search = (char *)buffer;
    int len = strlen(buffer);

    while ((search - buffer) < len) {
        if (!isspace(*search)) {
            *pos = search;
            return (search - buffer);
        }
        search++;
    }
    *pos = NULL;
    return -1;
}

/** Helper to check if there is some datas in a buffer
 * @param buffer
 */
static int check_arg (const char *buffer)
{
    char *pos;

    return (buffer != NULL && find_not_space(buffer, &pos) >= 0 && 
            strlen(pos) > 0);
}

/** Fake action routine
 */
static inline std_stype do_nothing (const char *bufargs) 
{
    return 0;
}

/** Fake action routine that returns DO_EXIT
 */
static std_stype do_exit (const char *bufargs) 
{
    return DO_EXIT;
}

/** Action routine called when an error occurs 
 *  and when help command is invoqued: displays help
 */
static std_stype do_cmd_error (const char *bufargs) 
{
    int i;
    struct commands *cmd = commands;

    while (cmd->cmd != NULL) {
        sdpl_info("%s ", cmd->cmd);
        for (i = 0; i < 3; i++) {
            if (i < cmd->arg_max)
                if (i < cmd->arg_min)
                    sdpl_info(" arg%1d  ", i);
                else
                    sdpl_info("[arg%1d] ", i);
            else
                sdpl_info("     ");
        }
        sdpl_info("\t: %s\n", cmd->comments);
        cmd++;
    }

    return 0;
}

/** Action routine that dump layer's main informations
 */
static std_stype do_print_sdpl_infos (const char *bufargs)
{
    dump_SDPL_infos(SDPL_map);
    return 0;
}

/** Action routine that display infos about one or all partitions
 */
static std_stype do_print_sdpl_part (const char *bufargs)
{
    std_type npart, part_min, part_max;

    if (check_arg(bufargs) == TRUE) {
        npart = (std_type)atoll(bufargs);
        if (npart >= NPARTS(SDPL_map)) {
            sdpl_error("Partition doesn't exist: " STD_ATTRIB 
                       " (" STD_ATTRIB ")\n", npart, NPARTS(SDPL_map));
            return -1;
        }
        part_min = npart;
        part_max = npart + 1;
    } else {
        part_min = 0;
        part_max = NPARTS(SDPL_map);
    }
    
    for (npart = part_min; npart < part_max; npart++) {
        dump_part_infos(SDPL_map, npart);
    }
    return 0;
}

/** Action routine that displays the free blocs list
 */
static std_stype do_print_free_blocs (const char *bufargs)
{
    dump_free_blocks(SDPL_map);
    return 0;
}

/** Action routine that reinitialize a layer
 */
static std_stype init_sdpl (const char *bufargs)
{
    std_type i, numparts;

    if (check_arg(bufargs) == TRUE) {
        switch (*bufargs) {
        case 'e':
            EMBEDDED(SDPL_map) = 1;
            break;
        default:
            sdpl_error("Only e (embedded) flag is allowed\n");
            return -1;
        }
    }
    if (SDPL_map->pmagic != NULL) {
        free(SDPL_map->pmagic);
        SDPL_map->pmagic = NULL;
    }
    if (SDPL_map->pflags != NULL) {
        free(SDPL_map->pflags);
        SDPL_map->pflags = NULL;
    }
    if (SDPL_map->poffset != NULL) {
        free(SDPL_map->poffset);
        SDPL_map->poffset = NULL;
    }
    if (SDPL_map->pnext != NULL) {
        free(SDPL_map->pnext);
        SDPL_map->pnext = NULL;
    }
    if (SDPL_map->pbloc != NULL) {
        for (i = 0; i < NPARTS(SDPL_map); i++) {
            if (SDPL_map->pbloc[i] != NULL) {
                free(SDPL_map->pbloc[i]);
            }
        }
        free(SDPL_map->pbloc);
        SDPL_map->pbloc = NULL;
    }
    if (SDPL_map->psize != NULL) {
        free(SDPL_map->psize);
        SDPL_map->psize = NULL;
    }
    if (SDPL_map->pentries != NULL) {
        free(SDPL_map->pentries);
        SDPL_map->pentries = NULL;
    }
    memset (SDPL_map->bflags, EMPTY_FILL, SDPL_map->SDPL_size * sizeof(__u16));
    NPARTS(SDPL_map) = 0;

    /* 
     * For embedded system, parameters are set so all information
     * can go in the boot-record
     */
    if (EMBEDDED(SDPL_map) == 1) {
        std_type old_ressize;

        SDPL_map->info_size = 16;
        /* We keep 1 ko for MBR + safety copy */
        old_ressize = SDPL_map->res_size;
        SDPL_map->res_size = 1024;
        SDPL_map->st_size += old_ressize - SDPL_map->res_size;
        SDPL_map->pentries_offset = 0x08;
        SDPL_map->bitmap_offset = 0x190;
        fix_SDPL_infos(SDPL_map);
        numparts = 4;
        for (i = 0; i < numparts; i++) {
            do_add_partition("0\n");
        }
    } else {
        std_type old_ressize;
        std_type pentry_size, pmap_size, bitmap_size;

        old_ressize = SDPL_map->res_size;
        SDPL_map->st_size += old_ressize;
        SDPL_map->res_size = get_ressize(SDPL_map->device_size, 
                                         max(64, SDPL_map->device_size >> 25), 
                                         SDPL_map->info_size, 0, &pentry_size, 
                                         &pmap_size, &bitmap_size);
        /* Let's consider we need one mapping bloc / sector */
#if 0
        pmap_size = (SDPL_map->st_size / 1024 / 1024 / 512) * 
            sizeof(union part_entry);
        pmap_size = ((pmap_size + 1023) / 1024) * 1024;
        /* Let's consider we can have 1 partition / Mbyte */
        pentry_size = max(4, SDPL_map->st_size / 1024) * 
            (SDPL_map->info_size / 8);
        pentry_size = ((pentry_size + 1023) / 1024) * 1024;
        SDPL_map->res_size = 1024 /* MBR + copy size */ +
            pmap_size /* Size for bloc mapping */ +
            pentry_size /* Size for partition entries */;
        bitmap_size = SDPL_map->SDPL_size / 8;
        bitmap_size = ((bitmap_size + 1023) / 1024) * 1024;
#endif
        SDPL_map->data_offset = SDPL_map->res_size;
        SDPL_map->st_size -= SDPL_map->res_size;
        fix_SDPL_infos(SDPL_map);
        SDPL_map->bitmap_offset = 0x400;
        SDPL_map->pentries_offset = 0x400 + bitmap_size;
        if (SDPL_map->info_size == 0)
            SDPL_map->info_size = sizeof(std_type) * 8;
    }

    return 0;
}

/** Action routine that changes the information size of the layer
 */
static std_stype change_info_size (const char *bufargs)
{
    std_type info_size;

    if (check_arg(bufargs) == FALSE) {
        do_cmd_error(NULL);
        return -1;
    }
    info_size = (std_type)atoll(bufargs);
    if (info_size == 16 || info_size == 32 
#if CONFIG_BLK_SDPL_64BITS
        || info_size == 64
#endif /* CONFIG_BLK_SDPL_64BITS */
        ) {
        SDPL_map->info_size = info_size;
        return 0;
    }
    sdpl_error("Unsupported info size: " STD_ATTRIB "\n", info_size);
    return -1;
}

/** Action routine that updates the layer's bloc size 
 */
static std_stype change_bloc_size (const char *bufargs)
{
    std_type bloc_size;
    std_type max_size = SDPL_map->st_size / 512;

    if (check_arg(bufargs) == FALSE) {
        do_cmd_error(NULL);
        return -1;
    }
    bloc_size = (std_type)atoll(bufargs);
    if (bloc_size <= max_size) {
        SDPL_map->bloc_size = bloc_size;
        SDPL_map->SDPL_size = max_size / SDPL_map->bloc_size;
        fix_bflags(SDPL_map);
        return 0;
    }
    sdpl_error("Bloc size cant be > device size (" STD_ATTRIB " <-> " 
               STD_ATTRIB ")\n", bloc_size * 512, SDPL_map->st_size);
    return -1;
}

/** Action routine that updates the total layer's size
 */
static std_stype change_layer_size (const char *bufargs)
{
    std_type layer_size;
    std_type max_size = SDPL_map->st_size / (512 * SDPL_map->bloc_size);

    if (check_arg(bufargs) == FALSE) {
        do_cmd_error(NULL);
        return -1;
    }
    layer_size = (std_type)atoll(bufargs);
    if (layer_size <= max_size) {
        SDPL_map->SDPL_size = layer_size;
        fix_bflags(SDPL_map);
        return 0;
    }
    sdpl_error("Bloc size cant be > device size (" STD_ATTRIB " <-> " 
               STD_ATTRIB ")\n", layer_size * SDPL_map->bloc_size * 512, 
               SDPL_map->st_size);
    return -1;
}
/** Action routine to change the partition infos offset
 */
#if 0
static std_stype change_part_offset (const char *bufargs)
{
    std_type offset;

    if (check_arg(bufargs) == FALSE) {
        sdpl_error("Missing offset\n");
    }
    offset = (std_type)atoll(bufargs);

    /* Should add "intelligent" routine to verify offset */
    SDPL_map->pentries_offset = offset;

    return 0;
}
#endif

/** Action routine that add a new partition to the layer
 */
static std_stype do_add_partition (const char *bufargs)
{
    std_type part_size;
    std_type max_size;

    if (EMBEDDED(SDPL_map) == 1)
        max_size = 40;
    else
        max_size = count_free_blocs(SDPL_map);
    if (check_arg(bufargs) == FALSE) {
        part_size = max_size;
    } else {
        part_size = (std_type)atoll(bufargs);
    }
    if (part_size <= max_size) {
        add_spartition(SDPL_map, part_size);
        return 0;
    }
    sdpl_error("Partition size cant be > free space (" STD_ATTRIB " <-> " 
               STD_ATTRIB ")\n", part_size, max_size);
    return -1;
}

static std_stype do_remove_partition (const char *bufargs)
{
    std_type part_num;

    if (check_arg(bufargs) == FALSE) {
        return -1;
    } else {
        part_num = (std_type)atoll(bufargs);
    }
    if (part_num <= NPARTS(SDPL_map)) {
        remove_spartition(SDPL_map, part_num);
        return 0;
    }
    sdpl_error("Partition " STD_ATTRIB " doesn't exists\n", part_num);
    return -1;
}

static std_stype do_resize_partition (const char *bufargs) 
{
    std_type numpart;
    std_stype resize;
    char *pos = (char *)bufargs;

    if (check_arg(pos) == FALSE) {
        sdpl_error("Should give the partition to resize\n");
        return -1;
    }
    numpart = atol(bufargs);
    if (numpart >= NPARTS(SDPL_map)) {
        sdpl_error("Bad partition\n");
        return -1;
    }
    while (*pos != '\0')
        pos++;
    pos++;
    while (isspace(*pos))
        pos++;
    /* if no size given, free all blocs */
    if (check_arg(pos) == FALSE) {
        resize = -SDPL_map->psize[numpart];
    } else {
        resize = (std_stype)atoll(pos);
    }
    if (resize < 0) {
        std_type first_to_free, i;

        if ((std_type)(-resize) > SDPL_map->psize[numpart]) {
            sdpl_error("Can't free more than part " STD_ATTRIB " length (" 
                       STD_ATTRIB " - " STD_ATTRIB ")\n", numpart, 
                       (-resize), SDPL_map->psize[numpart]);
            return -1;
        }

        first_to_free = SDPL_map->psize[numpart] + resize;
        for (i = first_to_free; i < SDPL_map->psize[numpart]; i++) {
            if (remove_bloc_from_part(SDPL_map, numpart, (-resize))) {
                sdpl_error("Error while removing bloc: " STD_ATTRIB "\n", i);
                return -1;
            }
        }
        return 0;
    }

    return add_blocs_to_part(SDPL_map, numpart, resize);
}

static std_stype do_mark_partition_bootable (const char *bufargs)
{
    std_type part_num;
    char *pos = (char *)bufargs;

    if (check_arg(pos) == FALSE) {
        sdpl_error("Should give the bloc to update\n");
        return -1;
    }
    part_num = atol(bufargs);
    if (part_num >= NPARTS(SDPL_map)) {
        sdpl_error("Bad part number\n");
        return -1;
    }
    mark_partition_bootable(SDPL_map, part_num);

    return 0;
}

static std_stype do_toggle_part_flag (const char *bufargs)
{
    std_type part_num;
    char *pos = (char *)bufargs;

    if (check_arg(pos) == FALSE) {
        sdpl_error("Should give the bloc to update\n");
        return -1;
    }
    part_num = atol(bufargs);
    if (part_num >= NPARTS(SDPL_map)) {
        sdpl_error("Bad part number\n");
        return -1;
    }
    while (*pos != '\0')
        pos++;
    pos++;
    while (isspace(*pos))
        pos++;
    /* if no flag given, toggle Read-Only flag */
    if (check_arg(pos) == FALSE)
        toggle_part_flag(SDPL_map, part_num, -1);
    else
        toggle_part_flag(SDPL_map, part_num, atoll(pos));

    return 0;
}

static std_stype do_toggle_bloc_flag (const char *bufargs)
{
    std_type bloc_num;
    char *pos = (char *)bufargs;

    if (check_arg(pos) == FALSE) {
        sdpl_error("Should give the bloc to update\n");
        return -1;
    }
    bloc_num = atol(bufargs);
    if (bloc_num >= SDPL_map->SDPL_size) {
        sdpl_error("Bad bloc number\n");
        return -1;
    }
    while (*pos != '\0')
        pos++;
    pos++;
    while (isspace(*pos))
        pos++;
    if (check_arg(pos) == FALSE)
        toggle_bloc_flag(SDPL_map, bloc_num, -1);
    else
        toggle_bloc_flag(SDPL_map, bloc_num, atoll(pos));

    return 0;
}

/** Action routine that commits changes on the physical device
 */
static std_stype do_save_sdpl (const char *bufargs) 
{
    save_sdpl(SDPL_map, bdev, &sector, &cur_sect);
    return 0;
}

/** Main editor interractivity routine
 */
void interract (struct SDPL_map *pSDPL_map, struct block_device *bloc_dev, 
                Sector sect)
{
    char cmd_buffer[MAX_CMD];
    struct commands *cmd;
    char *first, *pos;
    int cmd_len;
    int i;
    int ret;
    char try_cmd;

    bdev = bloc_dev;
    SDPL_map = pSDPL_map;
    sector = sect;

    while (1) {
    do_loop:
        fgets(cmd_buffer, MAX_CMD, stdin);

        cmd_len = strlen(cmd_buffer);
        i = find_not_space(cmd_buffer, &first);
        /* All spaces -> next one */
        if (i == -1) {
            do_nothing(cmd_buffer);
            continue;
        }

        pos = first;
        cmd_len = strlen(first);
        while ((pos - cmd_buffer) < cmd_len) {
            if (isspace(*pos))
                *pos = '\0';
            pos++;
        }

        pos = first;
        /* All commands are one letter long... */
        if (strlen(pos) != 1) {
            do_cmd_error(NULL);
            continue;
        }
        try_cmd = tolower(*pos);
        cmd = commands;
        while (cmd->cmd != NULL) {
            i = 0;
            while (cmd->cmd[i] != '\0') {
                if (try_cmd == cmd->cmd[i]) {
                    find_not_space(first + 2, &pos);
                    if ((ret = cmd->exec(pos)) < 0)
                        do_cmd_error(NULL);
                    if (ret == DO_EXIT)
                        return;
                    goto do_loop;
                }
                i++;
            }
            cmd++;
        }
        /* Bad command */
        if (cmd->cmd == NULL) {
            sdpl_error("Error: cmd not found (%c)\n", *pos);
            do_cmd_error(NULL);
            continue;
        }
    }
}
