/*
 * SDPL sdplresize.c: Utility to resize a SDPL partition
 *
 *   Copyright (C) 2002, 2003, 2004, Jocelyn MAYER <l_indien@magic.fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/sdpl.h>

static const unsigned char *progname;

static void usage (void)
{
    fprintf(stderr, "Usage: %s [-d device] [-n] [-v] [-h] partition size\n", 
            progname);
    fprintf(stderr, "       -n: simulate - no action\n");
    fprintf(stderr, "       -v: be verbose\n");
}

int main (int argc, char **argv)
{
    std_type blocsize = 0;
    std_type blocs, last;
    const unsigned char *devname = "/dev/sla";
    unsigned char *partition;
    int size, bloc_size;
    int wanted_size, wanted_blocs;
    int doc_fd, part_fd;
    int ret = -1;
    int flag_action = 1;
    int flag_verbose = 0;
    s8 c;

    progname = argv[0];
    while ((c = getopt(argc, argv, "d:nhv")) != -1) {
        switch(c) {
        case 'd':
            devname = optarg;
            break;
        case 'h':
            usage();
            return 1;
        case 'n':
            flag_action = 0;
            break;
        case 'v':
            flag_verbose++;
            break;
        default:
            fprintf(stderr, "Bad argument: %c\n", c);
            usage();
            return -1;
        }
    }
    if (optind + 2 != argc) {
        fprintf(stderr, "Bad number of arguments\n");
        usage();
        return -1;
    }
    partition = argv[optind];
    wanted_size = atoi(argv[optind + 1]) * 1024;

    doc_fd = open(devname, O_RDONLY);
    if (doc_fd < 0) {
        fprintf(stderr, "Unable to open device: %s\n", devname);
        return -1;
    }
    part_fd = open(partition, O_RDONLY);
    if (part_fd < 0) {
        fprintf(stderr, "Unable to open partition: %s\n", partition);
        close(doc_fd);
        return -1;
    }

    if ((ioctl(doc_fd, SDPL_GET_BLOCSIZE, &blocsize)) < 0) {
        fprintf(stderr, "Unable to retrieve bloc size\n");
        goto end;
    }
    bloc_size = (int)blocsize * 512;
    if (ioctl(part_fd, BLKGETSIZE, &size) < 0 || size < 0) {
        fprintf(stderr, "Unbale to get partition size\n");
        goto end;
    }
    size *= 512;

    wanted_blocs = wanted_size / bloc_size;
    if ((wanted_size - (wanted_blocs * bloc_size)) != 0)
        wanted_blocs++;
    wanted_blocs -= (size / bloc_size);
    if (flag_verbose > 0) {
        fprintf(stderr, "Orig size: %d - wanted: %d - diff: %d\n"
                "blocsize: " STD_ATTRIB " - bloc org: %d - new_blocs: %d " 
                "-> %d\n", size, wanted_size, wanted_size - size,
                bloc_size, size / bloc_size, wanted_size / bloc_size,
                wanted_blocs);
    }
    if (wanted_blocs > 0) {
        /* Beware: blocs is unisgned ! */
        blocs = wanted_blocs;
        if (flag_verbose > 0) {
            fprintf(stderr, "Going to add %d blocs to partition %s\n",
                    blocs, partition);
        }
        if (flag_action != 0) {
            if (ioctl(part_fd, SDPL_ADD_BLOCS, &blocs) < 0) {
                fprintf(stderr, "Unable to add blocs to partition %s\n", 
                        partition);
                goto end;
            }
        }
    } else if (wanted_blocs < 0) {
        /* Size is supposed to match an entire number of blocs */
        last = (size / bloc_size) - 1;
        while (wanted_blocs < 0) {
            if (flag_verbose > 0) {
                fprintf(stderr, "Going to remove bloc %d from partition %s\n",
                        last, partition);
            }
            if (flag_action != 0) {
                if (ioctl(part_fd, SDPL_FREE_BLOC, &last) < 0) {
                    fprintf(stderr, "Unable to remove bloc from %s (%d)\n",
                            partition, blocs);
                    goto end;
                }
            }
            last--;
            wanted_blocs++;
        }
    } else {
        if (flag_verbose > 0) {
            fprintf(stderr, "Nothing to do for resize of %s\n", partition);
        }
    }
    
    ret = 0;

 end:
    close(part_fd);
    close(doc_fd);

    return ret;
}
