/*
 * map-ctrl.c
 *
 * This program is a control program for the block device remapper
 * (bd_remap) Linux kernel module.
 */
/*
 * Copyright (C) 2003-2004 by Jocelyn Mayer <l_indien@magic.fr>
 * distributed under the GNU Public License
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* History:
 * 2003-01-27: Created
 */

/* TODO:
 * - Specify device instead of mount point
 */

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/ioctl.h>
#include <dirent.h>

#include <linux/types.h>
#include <linux/fs.h>

#include "bd_remap.h"
#include "lib_remap.h"

static void usage (void)
{
    fprintf(stderr, "Usage:\tmap-ctrl [-b bsize] [-x exclude] [-f excludeff] "
            "[-i inval] [-n noremap] [-X] [-u] [-v] [-h] [-d device] "
            "mount-point\n");
    fprintf(stderr, "map-ctrl -u device\n");
    fprintf(stderr, "\tthis will remap the given device in memory,\n");
    fprintf(stderr, "\t-b: define the memory-mapped bloc size\n");
    fprintf(stderr, "\t-x: the given file/directory will be replaced "
            "with empty blocs\n");
    fprintf(stderr, "\t-f: the given file/directory will be replaced "
            "with blocs filled with 0xFF\n");
    fprintf(stderr, "\t-i: the given file/directory will be replaced "
            "with invalid blocs\n");
    fprintf(stderr, "\t-n: the given file/directory won't be remapped\n");
    fprintf(stderr, "\t-X: simulate: do no kernel call\n");
    fprintf(stderr, "\t-u: used to unmap a device\n");
    fprintf(stderr, "\t-v: increase verbosity\n");
    fprintf(stderr, "\t-h: print this help\n");
    fprintf(stderr, "\t-d: use control device 'device' "
            "instead of /dev/remap\n");
    fprintf(stderr, "\t mount point is the mount point of the device to be "
            "remapped in memory\n");

    fprintf(stderr, "\tAny number of excluded or invalidated files or "
            "directories can be given\n");
    fprintf(stderr, "\tof course, those file will consume no memory\n");
}

__attribute__((__noreturn__))
int main (int argc, char **argv)
{
    struct stat st;
    file_t *special;
    unsigned char *mp, *chrdev;
    unsigned int nb_special, blocsize;
    int type, unmap, simulate, verbose;
    int minor;
    int c;
    int ret;

    unmap = 0;
    simulate = 0;
    verbose = 0;
    mp = NULL;
    ret = 1;
    blocsize = REMAP_BLOC_SIZE;
    special = NULL;
    nb_special = 0;
    chrdev = NULL;
    minor = -1;
    while ((c = getopt(argc, argv, "hx:i:f:n:ub:d:Xv")) >= 0) {
        switch (c) {
        case 'h':
            usage();
            ret = 0;
            goto end;

        case 'u':
            unmap = 1;
            break;

        case 'X':
            simulate = 1;
            break;

        case 'v':
            verbose++;
            break;

        case 'b':
            blocsize = strtoul(argv[optind - 1], NULL, 0);
            break;

        case 'd':
            chrdev = optarg;
            break;

        case 'x':
            type = REMAP_EMPTY;
            goto treat_file;

        case 'f':
            type = REMAP_EMPTYFF;
            goto treat_file;

        case 'n':
            type = REMAP_NOREMAP;
            goto treat_file;

        case 'i':
            type = REMAP_DISCARD;
            goto treat_file;

        default:
            fprintf(stderr, "Unknown argument: %c\n", c);
            usage();
            goto end;

        treat_file:
            if (bdrm_add_file(&special, &nb_special,
                              optarg, type, verbose) < 0)
                goto end;
        }
    }
    if (optind != (argc - 1)) {
        fprintf(stderr, "Bad number of arguments\n");
        usage();
        goto end;
    }
    mp = strdup(argv[optind]);
    if (mp == NULL) {
        fprintf(stderr, "Memory error\n");
        goto end;
    }
    /* Initialize remap library */
    bdrm_register_error_stream(stderr);
    if (verbose == 0)
        bdrm_register_msg_stream(NULL);
    else
        bdrm_register_msg_stream(stdout);
    if (lstat(mp, &st) < 0) {
        fprintf(stderr, "can't stat: %s\n", mp);
        goto end;
    }
    if (unmap == 1) {
        /* Unmap request */
        ret = bdrm_do_unmap(chrdev, st.st_dev, simulate);
        goto end;
    }

    if (bdrm_do_map(chrdev, &special, &nb_special, mp, blocsize,  verbose,
                    simulate, &minor) < 0)
        goto end;
    fprintf(stderr, "Device has been remapped to minor %d\n", minor);
    ret = 0;

end:        
    free(mp);
    bdrm_free_list(&special, nb_special);
        
    exit(ret);
}
